# validated: 2017-12-26 DV f9bece2ffbf7 edu/wpi/first/wpilibj/AnalogGyro.java
#----------------------------------------------------------------------------
# Copyright (c) 2008-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .analoginput import AnalogInput
from .gyrobase import GyroBase
from .interfaces import PIDSource

__all__ = ["AnalogGyro"]

class AnalogGyro(GyroBase):
    """Interface to a gyro device via an :class:`.AnalogInput`
    
    Use a rate gyro to return the robots heading relative to a starting
    position.  The Gyro class tracks the robots heading based on the starting
    position. As the robot rotates the new heading is computed by integrating
    the rate of rotation returned by the sensor. When the class is
    instantiated, it does a short calibration routine where it samples the
    gyro while at rest to determine the default offset. This is subtracted
    from each sample to determine the heading.

    .. not_implemented: initGyro
    """

    kOversampleBits = 10
    kAverageBits = 0
    kSamplesPerSecond = 50.0
    kCalibrationSampleTime = 5.0
    kDefaultVoltsPerDegreePerSecond = 0.007
    
    PIDSourceType = PIDSource.PIDSourceType
    
    def __init__(self, channel, center=None, offset=None):
        """Gyro constructor.

        Also initializes the gyro. Calibrate the gyro by running for a number
        of samples and computing the center value. Then use the
        center value as the Accumulator center value for subsequent
        measurements. It's important to make sure that the robot is not
        moving while the centering calculations are in progress, this is
        typically done when the robot is first turned on while it's sitting
        at rest before the competition starts.

        :param channel: The analog channel index or AnalogInput object that
            the gyro is connected to. Gyros can only be used on on-board channels 0-1.
        :param center: Preset uncalibrated value to use as the accumulator center value
        :type center: int
        :param offset: Preset uncalibrated value to use as the gyro offset
        :type offset: float
        """
        super().__init__()
        
        if not hasattr(channel, "initAccumulator"):
            channel = AnalogInput(channel)
            self.channelAllocated = True
            self.addChild(channel)
        else:
            self.channelAllocated = False
            
        self.analog = channel

        self.gyroHandle = hal.initializeAnalogGyro(self.analog.port)
    
        self.setDeadband(0.0)

        hal.setupAnalogGyro(self.gyroHandle)
        
        hal.report(hal.UsageReporting.kResourceType_Gyro, self.analog.getChannel())
        self.setName("AnalogGyro", self.analog.getChannel())
        
        if center is None or offset is None:
            self.calibrate()
        else:
            hal.setAnalogGyroParameters(self.gyroHandle, self.kDefaultVoltsPerDegreePerSecond, offset, center)
            self.reset()
        
    def calibrate(self):
        """:see: :meth:`.Gyro.calibrate`"""
        hal.calibrateAnalogGyro(self.gyroHandle)
        
    def reset(self):
        """:see: :meth:`.Gyro.reset`"""
        hal.resetAnalogGyro(self.gyroHandle)

    def free(self):
        """:see: :meth:`.Gyro.free`"""
        super().free()
        if self.analog is not None and self.channelAllocated:
            self.analog.free()
            self.analog = None
        hal.freeAnalogGyro(self.gyroHandle)
        self.gyroHandle = 0
        
    def getAngle(self):
        """:see: :meth:`.Gyro.getAngle`"""
    
        if self.analog is None:
            return 0.0
        
        return hal.getAnalogGyroAngle(self.gyroHandle)

    def getRate(self):
        """:see: :meth:`.Gyro.getRate`"""
        if self.analog is None:
            return 0.0
        else:
            return hal.getAnalogGyroRate(self.gyroHandle)

    def getOffset(self):
        """Return the gyro offset value set during calibration to
        use as a future preset
        
        :returns: the current offset value
        """
        return hal.getAnalogGyroOffset(self.gyroHandle)
    
    def getCenter(self):
        """Return the gyro center value set during calibration to
        use as a future preset
        
        :returns: the current center value
        """
        return hal.getAnalogGyroCenter(self.gyroHandle)

    def setSensitivity(self, voltsPerDegreePerSecond):
        """Set the gyro sensitivity. This takes the number of
        volts/degree/second sensitivity of the gyro and uses it in subsequent
        calculations to allow the code to work with multiple gyros. This value
        is typically found in the gyro datasheet.

        :param voltsPerDegreePerSecond:
            The sensitivity in Volts/degree/second
        :type  voltsPerDegreePerSecond: float
        """
        hal.setAnalogGyroVoltsPerDegreePerSecond(self.gyroHandle, voltsPerDegreePerSecond)

    def setDeadband(self, volts):
        """Set the size of the neutral zone.  Any voltage from the gyro less
        than this amount from the center is considered stationary.  Setting a
        deadband will decrease the amount of drift when the gyro isn't
        rotating, but will make it less accurate.

        :param volts: The size of the deadband in volts
        :type  volts: float
        """
        if self.analog is None:
            return
        
        hal.setAnalogGyroDeadband(self.gyroHandle, volts)




# validated: 2017-12-27 TW f9bece2ffbf7 edu/wpi/first/wpilibj/AnalogOutput.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .resource import Resource
from .sensorbase import SensorBase

__all__ = ["AnalogOutput"]

def _freeAnalogOutput(port):
    hal.freeAnalogOutputPort(port)

class AnalogOutput(SensorBase):
    """Analog output"""

    channels = Resource(SensorBase.kAnalogOutputChannels)

    def __init__(self, channel):
        """Construct an analog output on a specified MXP channel.

        :param channel: The channel number to represent.
        """
        super().__init__()
        SensorBase.checkAnalogOutputChannel(channel)
        
        self.channel = channel

        port = hal.getPort(channel)
        self._port = hal.initializeAnalogOutputPort(port)

        self.setName("AnalogOutput", channel)
        hal.report(hal.UsageReporting.kResourceType_AnalogChannel,
                      channel, 1)
        
        self.__finalizer = weakref.finalize(self, _freeAnalogOutput, self._port)

    @property
    def port(self):
        if not self.__finalizer.alive:
            return None
        return self._port
    
    def free(self):
        """Channel destructor.
        """
        super().free()
        if self.channel is None:
            return
        AnalogOutput.channels.free(self.channel)
        self.__finalizer()
        self.channel = None

    def getChannel(self):
        """Get the channel of this AnalogOutput.
        """
        return self.channel

    def setVoltage(self, voltage):
        hal.setAnalogOutput(self.port, voltage)

    def getVoltage(self):
        return hal.getAnalogOutput(self.port)

    def initSendable(self, builder):
        builder.setSmartDashboardType("Analog Output")
        builder.addDoubleProperty("Value", self.getVoltage, self.setVoltage)




# validated: 2017-11-23 TW e1195e8b9dab edu/wpi/first/wpilibj/DigitalSource.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal
import weakref

from .resource import Resource
from .sensorbase import SensorBase
from .interruptablesensorbase import InterruptableSensorBase

__all__ = ["DigitalSource"]


def _freeDigitalSource(handle):
    hal.freeDIOPort(handle)


class DigitalSource(InterruptableSensorBase):
    """DigitalSource Interface. The DigitalSource represents all the possible
    inputs for a counter or a quadrature encoder. The source may be either a
    digital input or an analog input. If the caller just provides a channel,
    then a digital input will be constructed and freed when finished for the
    source. The source can either be a digital input or analog trigger but
    not both.
    """

    def isAnalogTrigger(self):
        raise NotImplementedError

    def getChannel(self):
        raise NotImplementedError



# validated: 2017-12-25 TW f9bece2ffbf7 edu/wpi/first/wpilibj/drive/RobotDriveBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from ..motorsafety import MotorSafety
from ..sendablebase import SendableBase

__all__ = ["RobotDriveBase"]


class RobotDriveBase(SendableBase, MotorSafety):
    """Common base class for drive platforms"""

    class MotorType:
        """The location of a motor on the robot for the purpose of driving."""

        #: Front left
        kFrontLeft = 0

        #: Front right
        kFrontRight = 1

        #: Rear left
        kRearLeft = 2

        #: Rear right
        kRearRight = 3

        #: Left
        kLeft = 0

        #: Right
        kRight = 1

        #: Back
        kBack = 2

    kDefaultDeadband = .02
    kDefaultMaxOutput = 1.0

    def __init__(self):
        SendableBase.__init__(self)
        MotorSafety.__init__(self)

        self.deadband = self.kDefaultDeadband
        self.maxOutput = self.kDefaultMaxOutput

        self.setSafetyEnabled(True)
        self.setName("RobotDriveBase")

    def setDeadband(self, deadband: float):
        """Change the value for deadband scaling.

        The default value is :const:`kDefaultDeadband`. Values smaller than the deadband are set to 0,
        while values larger than the deadband are scaled from 0.0 to 1.0. See :meth:`applyDeadband`.

        :param deadband: The deadband to set
        """
        self.deadband = deadband

    def setMaxOutput(self, maxOutput: float):
        """Configure the scaling factor for using drive methods with motor controllers in a mode
        other than PercentVbus or to limit the maximum output.

        The default value is :const:`kDefaultMaxOutput`.

        :param maxOutput: Multiplied with the output percentage computed by the drive functions.
        """
        self.maxOutput = maxOutput

    @staticmethod
    def limit(value):
        """Limit motor values to the -1.0 to +1.0 range."""
        if value > 1.0:
            return 1.0
        if value < -1.0:
            return -1.0
        return value

    @staticmethod
    def applyDeadband(value, deadband):
        """Returns 0.0 if the given value is within the specified range around zero. The remaining range
        between the deadband and 1.0 is scaled from 0.0 to 1.0.

        :param value: value to clip
        :param deadband: range around zero
        """
        if abs(value) > deadband:
            if value < 0.0:
                return (value - deadband) / (1.0 - deadband)
            else:
                return (value + deadband) / (1.0 - deadband)
        return 0.0

    @staticmethod
    def normalize(wheelSpeeds):
        """Normalize all wheel speeds if the magnitude of any wheel is greater
        than 1.0.

        :param wheelSpeeds: Iterable of wheelspeeds to normalize
        """
        maxMagnitude = max(abs(x) for x in wheelSpeeds)
        if maxMagnitude > 1.0:
            for i in range(len(wheelSpeeds)):
                wheelSpeeds[i] = wheelSpeeds[i] / maxMagnitude



# validated: 2017-12-22 TW 3c3a448d4726 edu/wpi/first/wpilibj/SampleRobot.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import logging
import warnings

import hal
from .livewindow import LiveWindow
from .robotbase import RobotBase
from .timer import Timer

__all__ = ["SampleRobot"]


class SampleRobot(RobotBase):
    """A simple robot base class that knows the standard FRC competition
    states (disabled, autonomous, or operator controlled).

    You can build a simple robot program off of this by overriding the
    :meth:`robotinit`, :meth:`disabled`, :meth:`autonomous` and
    :meth:`operatorControl` methods. The :meth:`startCompetition` method
    will call these methods (sometimes repeatedly) depending on the state
    of the competition.

    Alternatively you can override the :meth:`robotMain` method and manage all
    aspects of the robot yourself (not recommended).
    
    .. warning::
        While it may look like a good choice to use for your code
        if you're inexperienced, don't. Unless you know what you
        are doing, complex code will be much more difficult under
        this system. Use :class:`.TimedRobot` or command based
        instead if you're new.

    .. deprecated:: 2018.0.0
    """

    #: A python logging object that you can use to send messages to the log. It
    #: is recommended to use this instead of print statements.
    logger = logging.getLogger("robot")

    def robotInit(self):
        """Robot-wide initialization code should go here.

        Users should override this method for default Robot-wide initialization
        which will be called when the robot is first powered on.  It will be
        called exactly 1 time.
        
        .. note:: It is simpler to override this function instead of defining
                  a constructor for your robot class
                  
        .. warning:: the Driver Station "Robot Code" light and FMS "Robot Ready"
                     indicators will be off until RobotInit() exits. Code in ``robotInit()`` that
                     waits for enable will cause the robot to never indicate that the code is
                     ready, causing the robot to be bypassed in a match.
        """
        self.logger.info("Default robotInit() method running, consider providing your own")

    def disabled(self):
        """Disabled should go here.
        Users should overload this method to run code that should run while
        the field is disabled.

        Called once each time the robot enters the disabled state.
        """
        self.logger.info("Default disabled() method running, consider providing your own")

    def autonomous(self):
        """Autonomous should go here.
        Users should add autonomous code to this method that should run while
        the field is in the autonomous period.

        Called once each time the robot enters the autonomous state.
        """
        self.logger.info("Default autonomous() method running, consider providing your own")

    def operatorControl(self):
        """Operator control (tele-operated) code should go here.
        Users should add Operator Control code to this method that should run
        while the field is in the Operator Control (tele-operated) period.

        Called once each time the robot enters the operator-controlled state.
        """
        self.logger.warning("Default operatorControl() method running, consider providing your own")

    def test(self):
        """Test code should go here.
        Users should add test code to this method that should run while the
        robot is in test mode.
        """
        self.logger.info("Default test() method running, consider providing your own")

    def robotMain(self):
        """Robot main program for free-form programs.

        This should be overridden by user subclasses if the intent is to not
        use the autonomous() and operatorControl() methods. In that case, the
        program is responsible for sensing when to run the autonomous and
        operator control functions in their program.

        This method will be called immediately after the constructor is
        called. If it has not been overridden by a user subclass (i.e. the
        default version runs), then the robotInit(), disabled(), autonomous()
        and operatorControl() methods will be called.
        
        If you override this function, you must call ``hal.HALNetworkCommunicationObserveUserProgramStarting()``
        to indicate that your robot is ready to be enabled, as it will not
        be called for you.
        
        .. warning:: Nobody actually wants to override this function. Neither
                     do you.
        
        """
        self._no_robot_main = True

    def startCompetition(self):
        """Start a competition.
        This code tracks the order of the field starting to ensure that
        everything happens in the right order. Repeatedly run the correct
        method, either Autonomous or OperatorControl when the robot is
        enabled. After running the correct method, wait for some state to
        change, either the other mode starts or the robot is disabled. Then
        go back and wait for the robot to be enabled again.
        """
        hal.report(hal.UsageReporting.kResourceType_Framework,
                   hal.UsageReporting.kFramework_Sample)

        self.robotInit()

        # Tell the DS that the robot is ready to be enabled
        hal.observeUserProgramStarting()

        self.robotMain()

        if hasattr(self, '_no_robot_main'):
            while True:
                if self.isDisabled():
                    self.ds.InDisabled(True)
                    self.disabled()
                    self.ds.InDisabled(False)
                    while self.isDisabled():
                        Timer.delay(0.01)
                elif self.isAutonomous():
                    self.ds.InAutonomous(True)
                    self.autonomous()
                    self.ds.InAutonomous(False)
                    while self.isAutonomous() and not self.isDisabled():
                        Timer.delay(0.01)
                elif self.isTest():
                    LiveWindow.setEnabled(True)
                    self.ds.InTest(True)
                    self.test()
                    self.ds.InTest(False)
                    while self.isTest() and self.isEnabled():
                        Timer.delay(0.01)
                    LiveWindow.setEnabled(False)
                else:
                    self.ds.InOperatorControl(True)
                    self.operatorControl()
                    self.ds.InOperatorControl(False)
                    while self.isOperatorControl() and not self.isDisabled():
                        Timer.delay(0.01)



# validated: 2017-12-12 EN f9bece2ffbf7 edu/wpi/first/wpilibj/DoubleSolenoid.java
#----------------------------------------------------------------------------
# Copyright (c) 2008-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import enum
import warnings
import weakref

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase
from .solenoidbase import SolenoidBase

__all__ = ["DoubleSolenoid"]

def _freeSolenoid(fwdHandle, revHandle):
    hal.freeSolenoidPort(fwdHandle)
    hal.freeSolenoidPort(revHandle)

class DoubleSolenoid(SolenoidBase):
    """Controls 2 channels of high voltage Digital Output on the PCM.

    The DoubleSolenoid class is typically used for pneumatics solenoids that
    have two positions controlled by two separate channels.
    """

    class Value(enum.IntEnum):
        """Possible values for a DoubleSolenoid."""
        kOff = 0
        kForward = 1
        kReverse = 2

    def __init__(self, *args, **kwargs):
        """Constructor.

        Arguments can be supplied as positional or keyword.  Acceptable
        positional argument combinations are:
        
        - forwardChannel, reverseChannel
        - moduleNumber, forwardChannel, reverseChannel

        Alternatively, the above names can be used as keyword arguments.

        :param moduleNumber: The module number of the solenoid module to use.
        :param forwardChannel: The forward channel number on the module to control (0..7)
        :param reverseChannel: The reverse channel number on the module to control  (0..7)
        """
        # keyword arguments
        forwardChannel = kwargs.pop("forwardChannel", None)
        reverseChannel = kwargs.pop("reverseChannel", None)
        moduleNumber = kwargs.pop("moduleNumber", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 2:
            forwardChannel, reverseChannel = args
        elif len(args) == 3:
            moduleNumber, forwardChannel, reverseChannel = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        if moduleNumber is None:
            moduleNumber = SensorBase.getDefaultSolenoidModule()
        if forwardChannel is None:
            raise ValueError("must specify forward channel")
        if reverseChannel is None:
            raise ValueError("must specify reverse channel")

        super().__init__(moduleNumber)
        
        self.valueEntry = None
        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.checkSolenoidChannel(forwardChannel)
        SensorBase.checkSolenoidChannel(reverseChannel)

        portHandle = hal.getPortWithModule(moduleNumber, forwardChannel)
        self.forwardHandle = hal.initializeSolenoidPort(portHandle)
        
        try:
            portHandle = hal.getPortWithModule(moduleNumber, reverseChannel)
            self.reverseHandle = hal.initializeSolenoidPort(portHandle)
        except Exception:
            # free the forward handle on exception, then rethrow
            hal.freeSolenoidPort(self.forwardHandle)
            self.forwardHandle = None
            self.reverseHandle = None
            raise

        self.forwardMask = 1 << forwardChannel
        self.reverseMask = 1 << reverseChannel

        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        hal.report(hal.UsageReporting.kResourceType_Solenoid,
                      forwardChannel, moduleNumber)
        hal.report(hal.UsageReporting.kResourceType_Solenoid,
                      reverseChannel, moduleNumber)

        self.setName("DoubleSolenoid", moduleNumber, forwardChannel)
        
        self.__finalizer = weakref.finalize(self, _freeSolenoid,
                                            self.forwardHandle, self.reverseHandle)

    def free(self):
        """Mark the solenoid as freed."""
        super().free()
        self.__finalizer()
        self.forwardHandle = None
        self.reverseHandle = None

    def set(self, value):
        """Set the value of a solenoid.

        :param value: The value to set (Off, Forward, Reverse)
        :type  value: :class:`DoubleSolenoid.Value`
        """

        if value == self.Value.kOff:
            hal.setSolenoid(self.forwardHandle, False)
            hal.setSolenoid(self.reverseHandle, False)
        elif value == self.Value.kForward:
            hal.setSolenoid(self.reverseHandle, False)
            hal.setSolenoid(self.forwardHandle, True)
        elif value == self.Value.kReverse:
            hal.setSolenoid(self.forwardHandle, False)
            hal.setSolenoid(self.reverseHandle, True)
        else:
            raise ValueError("Invalid argument '%s'" % value)

    def get(self):
        """Read the current value of the solenoid.

        :returns: The current value of the solenoid.
        :rtype: :class:`DoubleSolenoid.Value`
        """
        if hal.getSolenoid(self.forwardHandle):
            return self.Value.kForward
        if hal.getSolenoid(self.reverseHandle):
            return self.Value.kReverse
        return self.Value.kOff

    def isFwdSolenoidBlackListed(self):
        """
        Check if the forward solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        blacklist = self.getPCMSolenoidBlackList()

        return (blacklist & self.forwardMask) != 0

    def isRevSolenoidBlackListed(self):
        """
        Check if the reverse solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        blacklist = self.getPCMSolenoidBlackList()

        return blacklist & (1 << self.reverseMask) != 0

    def initSendable(self, builder):
        builder.setSmartDashboardType("Double Solenoid")
        builder.setSafeState(lambda: self.set(self.Value.kOff))
        builder.addStringProperty("Value", lambda: self.get().name[1:], self._valueChanged)

    def _valueChanged(self, value):
        if value == "Reverse":
            self.set(self.Value.kReverse)
        elif value == "Forward":
            self.set(self.Value.kForward)
        else:
            self.set(self.Value.kOff)



# validated: 2017-12-21 DV de134a5c608d edu/wpi/first/wpilibj/NamedSendable.java
#----------------------------------------------------------------------------
# Copyright (c) 2012-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import warnings

from ..sendable import Sendable

__all__ = ["NamedSendable"]

class NamedSendable(Sendable):
    """The interface for sendable objects that gives the sendable a default
    name in the Smart Dashboard.

    .. deprecated:: 2018.0
        Use :class:`.Sendable` directly instead.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        warnings.warn('NamedSendable is deprecated, use Sendable directly instead',
                      DeprecationWarning, stacklevel=2)

    def getName(self) -> str:
        """The name of the subtable.

        :returns: The name of the subtable of SmartDashboard that the
                  :class:`.Sendable` object will use
        """
        raise NotImplementedError

    def setName(self, name):
        pass

    def getSubsystem(self):
        return ""

    def setSubsystem(self, subsystem):
        pass

    def initSendable(self, builder):
        pass



# validated: 2017-12-12 EN f9bece2ffbf7 edu/wpi/first/wpilibj/Solenoid.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref
import warnings

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase
from .solenoidbase import SolenoidBase

__all__ = ["Solenoid"]

def _freeSolenoid(solenoidHandle):
    hal.freeSolenoidPort(solenoidHandle)

class Solenoid(SolenoidBase):
    """Solenoid class for running high voltage Digital Output.

    The Solenoid class is typically used for pneumatic solenoids, but could
    be used for any device within the current spec of the PCM.
    
    .. not_implemented: initSolenoid
    """

    def __init__(self, *args, **kwargs):
        """Constructor.

        Arguments can be supplied as positional or keyword.  Acceptable
        positional argument combinations are:
        
        - channel
        - moduleNumber, channel

        Alternatively, the above names can be used as keyword arguments.

        :param moduleNumber: The CAN ID of the PCM the solenoid is attached to
        :type moduleNumber: int
        :param channel: The channel on the PCM to control (0..7)
        :type channel: int
        """
        # keyword arguments
        channel = kwargs.pop("channel", None)
        moduleNumber = kwargs.pop("moduleNumber", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 1:
            channel = args[0]
        elif len(args) == 2:
            moduleNumber, channel = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        if moduleNumber is None:
            moduleNumber = SensorBase.getDefaultSolenoidModule()
        if channel is None:
            raise ValueError("must specify channel")

        super().__init__(moduleNumber)
        self.channel = channel

        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.checkSolenoidChannel(channel)
        
        portHandle = hal.getPortWithModule(moduleNumber, channel)
        self._solenoidHandle = hal.initializeSolenoidPort(portHandle)

        hal.report(hal.UsageReporting.kResourceType_Solenoid, channel,
                   moduleNumber)
        self.setName("Solenoid", self.moduleNumber, self.channel)
        
        self.__finalizer = weakref.finalize(self, _freeSolenoid, self._solenoidHandle)
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)
        
    @property
    def solenoidHandle(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use channel after free() has been called")
        return self._solenoidHandle

    def free(self):
        """Mark the solenoid as freed."""
        super().free()

        self.__finalizer()
        self._solenoidHandle = None
        

    def set(self, on):
        """Set the value of a solenoid.

        :param on: True will turn the solenoid output on. False will turn the solenoid output off.
        :type on: bool
        """
        hal.setSolenoid(self.solenoidHandle, on)

    def get(self):
        """Read the current value of the solenoid.

        :returns: True if the solenoid output is on or false if the solenoid output is off.
        :rtype: bool
        """
        return hal.getSolenoid(self.solenoidHandle)

    def isBlackListed(self):
        """
        Check if the solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`.SolenoidBase.clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        value = self.getPCMSolenoidBlackList() & (1 << self.channel)
        return value != 0

    def setPulseDuration(self, durationSeconds):
        """
        Set the pulse duration in the PCM. This is used in conjunction with
        the startPulse method to allow the PCM to control the timing of a pulse.
        The timing can be controlled in 0.01 second increments.

        see :meth:`startPulse`

        :param durationSeconds: The duration of the pulse, from 0.01 to 2.55 seconds.
        """
        duration_ms = int(durationSeconds * 1000)
        hal.setOneShotDuration(self.solenoidHandle, duration_ms)

    def startPulse(self):
        """
        Trigger the PCM to generate a pulse of the duration set in
        setPulseDuration.

        see :meth:`setPulseDuration`
        """
        hal.fireOneShot(self.solenoidHandle)

    def initSendable(self, builder):
        builder.setSmartDashboardType("Solenoid")
        builder.setSafeState(lambda: self.set(False))
        builder.addBooleanProperty("Value", self.get, self.set)



# validated: 2017-11-21 EN b65447b6f5a8 edu/wpi/first/wpilibj/CounterBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["CounterBase"]

class CounterBase:
    """Interface for counting the number of ticks on a digital input channel.
    Encoders, Gear tooth sensors, and counters should all subclass this so it
    can be used to build more advanced classes for control and driving.

    All counters will immediately start counting - :meth:`reset` them if you
    need them to be zeroed before use.
    """

    class EncodingType:
        """The number of edges for the counterbase to increment or decrement on"""
        
        #: Count only the rising edge
        k1X = 0
        
        #: Count both the rising and falling edge
        k2X = 1
        
        #: Count rising and falling on both channels
        k4X = 2

    def get(self):
        """Get the count
        
        :returns: the count
        """
        raise NotImplementedError

    def reset(self):
        """Reset the count to zero"""
        raise NotImplementedError

    def getPeriod(self):
        """Get the time between the last two edges counted
        
        :returns: the time between the last two ticks in seconds
        """
        raise NotImplementedError

    def setMaxPeriod(self, maxPeriod):
        """Set the maximum time between edges to be considered stalled
        
        :param maxPeriod: the maximum period in seconds
        """
        raise NotImplementedError

    def getStopped(self):
        """Determine if the counter is not moving
        
        :returns: True if the counter has not changed for the max period
        """
        raise NotImplementedError

    def getDirection(self):
        """Determine which direction the counter is going
        
        :returns: True for one direction, False for the other
        """
        raise NotImplementedError



# validated: 2017-02-19 DS 12f759860eea edu/wpi/first/wpilibj/I2C.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import warnings
import weakref

__all__ = ["I2C"]

def _freeI2C(port):
    hal.closeI2C(port)

class I2C:
    """I2C bus interface class.

    This class is intended to be used by sensor (and other I2C device) drivers.
    It probably should not be used directly.
    
    Example usage::
    
        i2c = wpilib.I2C(wpilib.I2C.Port.kOnboard, 4)
        
        # Write bytes 'text', and receive 4 bytes in data
        data = i2c.transaction(b'text', 4)
    """
    class Port:
        kOnboard = 0
        kMXP = 1

    def __init__(self, port, deviceAddress, simPort=None):
        """Constructor.

        :param port: The I2C port the device is connected to.
        :type port: :class:`.I2C.Port`
        :param deviceAddress: The address of the device on the I2C bus.
        :param simPort: This must be an object that implements all of
                        the i2c* functions from hal_impl that you use.
                        See ``test_i2c.py`` for an example.
        """
        if port not in [self.Port.kOnboard, self.Port.kMXP]:
            raise ValueError("Invalid value '%s' for I2C port" % port)
        
        if hal.HALIsSimulation():
            if simPort is None:
                # If you want more functionality, implement your own mock
                from hal_impl.i2c_helpers import I2CSimBase
                simPort = I2CSimBase()
                
                msg = "Using stub simulator for I2C port %s" % port
                warnings.warn(msg)
            
            # Just check for basic functionality
            assert hasattr(simPort, 'initializeI2C')
            assert hasattr(simPort, 'closeI2C')
            
            self._port = (simPort, port)
        else:
            self._port = port
            
        self.deviceAddress = deviceAddress

        hal.initializeI2C(self._port)
        self.__finalizer = weakref.finalize(self, _freeI2C, self._port)

        hal.report(hal.UsageReporting.kResourceType_I2C, deviceAddress)
    
    @property
    def port(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use i2c port after free() has been called")
        return self._port
    
    def free(self):
        self.__finalizer()

    def transaction(self, dataToSend, receiveSize):
        """Generic transaction.

        This is a lower-level interface to the I2C hardware giving you more
        control over each transaction.

        :param dataToSend: Buffer of data to send as part of the transaction.
        :type dataToSend: iterable of bytes
        :param receiveSize: Number of bytes to read from the device.
        :type receiveSize: int
        :returns: Data received from the device.
        :rtype: iterable of bytes
        """
        return hal.transactionI2C(self.port, self.deviceAddress,
                                  dataToSend, receiveSize)

    def addressOnly(self):
        """Attempt to address a device on the I2C bus.

        This allows you to figure out if there is a device on the I2C bus that
        responds to the address specified in the constructor.

        :returns: Transfer Aborted... False for success, True for aborted.
        """
        try:
            self.transaction([], 0)
        except IOError:
            return True
        return False

    def write(self, registerAddress, data):
        """Execute a write transaction with the device.

        Write a single byte to a register on a device and wait until the
        transaction is complete.

        :param registerAddress:
            The address of the register on the device to be written.
        :param data: The byte to write to the register on the device.
        :returns: Transfer Aborted... False for success, True for aborted.
        """
        try:
            hal.writeI2C(self.port, self.deviceAddress, [registerAddress, data])
        except IOError:
            return True
        return False

    def writeBulk(self, data):
        """Execute a write transaction with the device.

        Write multiple bytes to a register on a device and wait until the
        transaction is complete.

        :param data: The data to write to the device.
        :type data: iterable of bytes
        :returns: Transfer Aborted... False for success, True for aborted.
        
        Usage::
        
            # send byte string
            failed = i2c.writeBulk(b'stuff')
            
            # send list of integers
            failed = i2c.write([0x01, 0x02])
        """
        try:
            hal.writeI2C(self.port, self.deviceAddress, data)
        except IOError:
            return True
        return False

    def read(self, registerAddress, count):
        """Execute a read transaction with the device.

        Read bytes from a device. Most I2C devices will auto-increment
        the register pointer internally allowing you to read
        consecutive registers on a device in a single transaction.

        :param registerAddress: The register to read first in the transaction.
        :param count: The number of bytes to read in the transaction.
        :returns: The data read from the device.
        :rtype: iterable of bytes
        """
        if count < 1:
            raise ValueError("count must be at least 1, %s given" % count)
        return self.transaction([registerAddress], count)

    def readOnly(self, count):
        """Execute a read only transaction with the device.

        Read bytes from a device. This method does not write any data
        to prompt the device.

        :param count: The number of bytes to read in the transaction.
        :returns: The data read from the device.
        :rtype: iterable of bytes
        """
        if count < 1:
            raise ValueError("count must be at least 1, %s given" % count)
        return hal.readI2C(self.port, self.deviceAddress, count)

    def verifySensor(self, registerAddress, expected):
        """Verify that a device's registers contain expected values.

        Most devices will have a set of registers that contain a known value
        that can be used to identify them. This allows an I2C device driver
        to easily verify that the device contains the expected value.

        The device must support and be configured to use register
        auto-increment.

        :param registerAddress:
            The base register to start reading from the device.
        :param expected: The values expected from the device.
        :returns: True if the sensor was verified to be connected
        """
        # TODO: Make use of all 7 read bytes
        count = len(expected)
        for i in range(0, count, 4):
            toRead = 4
            if (count - i) < 4:
                toRead = count - i
            # Read the chunk of data. Return False if the sensor does not
            # respond.
            try:
                deviceData = self.read(registerAddress, toRead)
            except IOError:
                return False

            for j in range(toRead):
                if deviceData[j] != expected[i + j]:
                    return False

            registerAddress += count

        return True



# validated: 2017-12-25 TW f9bece2ffbf7 edu/wpi/first/wpilibj/drive/MecanumDrive.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import math

import hal
from .robotdrivebase import RobotDriveBase
from .vector2d import Vector2d

__all__ = ["MecanumDrive"]


class MecanumDrive(RobotDriveBase):
    r"""A class for driving Mecanum drive platforms.

    Mecanum drives are rectangular with one wheel on each corner. Each wheel has rollers toed in
    45 degrees toward the front or back. When looking at the wheels from the top, the roller axles
    should form an X across the robot. Each drive() function provides different inverse kinematic
    relations for a Mecanum drive robot.

    Drive base diagram::

        \_______/
        \ |   | /
          |   |
        /_|___|_\
        /       \


    Each drive() function provides different inverse kinematic relations for a Mecanum drive
    robot. Motor outputs for the right side are negated, so motor direction inversion by the user is
    usually unnecessary.

    This library uses the NED axes convention (North-East-Down as external reference in the world
    frame): http://www.nuclearprojects.com/ins/images/axis_big.png.

    The positive X axis points ahead, the positive Y axis points right, and the positive Z axis
    points down. Rotations follow the right-hand rule, so clockwise rotation around the Z axis is
    positive.

    .. note:: RobotDrive porting guide:

        In MecanumDrive, the right side speed controllers are automatically inverted,
        while in RobotDrive, no speed controllers are automatically inverted.

        :meth:`driveCartesian` is equivalent to :meth:`.RobotDrive.mecanumDrive_Cartesian` if a deadband of 0
        is used, and the ``ySpeed`` and ``gyroAngle`` values are inverted compared to ``RobotDrive`` (i.e
        ``driveCartesian(xSpeed, -ySpeed, zRotation, -gyroAngle)``.

        :meth:`drivePolar` is equivalent to :meth:`.RobotDrive.mecanumPolar` if a deadband of 0 is used.
    """

    instances = 0

    def __init__(self, frontLeftMotor, rearLeftMotor, frontRightMotor, rearRightMotor):
        """Constructor for MecanumDrive.

        If motors need to be inverted, do so beforehand.
        Motor outputs for the right side are negated, so motor direction inversion
        by the user is usually unnecessary

        :param frontLeftMotor: Front Left Motor
        :param rearLeftMotor: Rear Left Motor
        :param frontRightMotor: Front Right Motor
        :param rearRightMotor: Rear Right Motor
        """
        super().__init__()

        self.frontLeftMotor = frontLeftMotor
        self.rearLeftMotor = rearLeftMotor
        self.frontRightMotor = frontRightMotor
        self.rearRightMotor = rearRightMotor

        self.addChild(self.frontLeftMotor)
        self.addChild(self.rearLeftMotor)
        self.addChild(self.frontRightMotor)
        self.addChild(self.rearRightMotor)
        MecanumDrive.instances += 1
        self.setName("MecanumDrive", self.instances)

        self.reported = False

    def driveCartesian(self, ySpeed, xSpeed, zRotation, gyroAngle=0.0):
        """Drive method for Mecanum platform.

        Angles are measured clockwise from the positive X axis. The robot's speed is independent
        from its angle or rotation rate.

        :param ySpeed: The robot's speed along the Y axis [-1.0..1.0]. Right is positive.
        :param xSpeed: The robot's speed along the X axis [-1.0..1.0]. Forward is positive.
        :param zRotation: The robot's rotation rate around the Z axis [-1.0..1.0]. Clockwise is positive.
        :param gyroAngle: The current angle reading from the gyro in degrees around the Z axis. Use
                          this to implement field-oriented controls.
        """
        if not self.reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       4,
                       hal.UsageReporting.kRobotDrive_MecanumCartesian)
            self.reported = True

        ySpeed = RobotDriveBase.limit(ySpeed)
        ySpeed = RobotDriveBase.applyDeadband(ySpeed, self.deadband)

        xSpeed = RobotDriveBase.limit(xSpeed)
        xSpeed = RobotDriveBase.applyDeadband(xSpeed, self.deadband)

        # Compensate for gyro angle
        input = Vector2d(ySpeed, xSpeed)
        input.rotate(gyroAngle)

        wheelSpeeds = [
            # Front Left
            input.x + input.y + zRotation,
            # Rear Left
            -input.x + input.y + zRotation,
            # Front Right
            input.x - input.y + zRotation,
            # Rear Right
            -input.x - input.y + zRotation
        ]

        RobotDriveBase.normalize(wheelSpeeds)

        wheelSpeeds = [speed * self.maxOutput for speed in wheelSpeeds]

        self.frontLeftMotor.set(wheelSpeeds[0])
        self.rearLeftMotor.set(wheelSpeeds[1])
        self.frontRightMotor.set(wheelSpeeds[2])
        self.rearRightMotor.set(wheelSpeeds[3])

        self.feed()

    def drivePolar(self, magnitude, angle, zRotation):
        """Drive method for Mecanum platform.

        Angles are measured counter-clockwise from straight ahead. The speed at which the robot
        drives (translation) is independent from its angle or rotation rate.

        :param magnitude: The robot's speed at a given angle [-1.0..1.0]. Forward is positive.
        :param angle: The angle around the Z axis at which the robot drives in degrees [-180..180].
        :param zRotation: The robot's rotation rate around the Z axis [-1.0..1.0]. Clockwise is
                          positive.
        """
        if not self.reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       4,
                       hal.UsageReporting.kRobotDrive_MecanumPolar)
            self.reported = True

        magnitude = RobotDriveBase.limit(magnitude) * math.sqrt(2)

        self.driveCartesian(magnitude * math.cos(math.radians(angle)),
                            magnitude * math.sin(math.radians(angle)),
                            zRotation, 0.0)

    def stopMotor(self):
        self.frontLeftMotor.stopMotor()
        self.rearLeftMotor.stopMotor()
        self.frontRightMotor.stopMotor()
        self.rearRightMotor.stopMotor()
        self.feed()

    def getDescription(self):
        return "Mecanum Drive"

    def initSendable(self, builder):
        builder.setSmartDashboardType("MecanumDrive")
        builder.addDoubleProperty("Front Left Motor Speed", self.frontLeftMotor.get, self.frontLeftMotor.set)
        builder.addDoubleProperty("Front Right Motor Speed", self.frontRightMotor.get, self.frontRightMotor.set)
        builder.addDoubleProperty("Rear Left Motor Speed", self.rearLeftMotor.get, self.rearLeftMotor.set)
        builder.addDoubleProperty("Rear Right Motor Speed", self.rearRightMotor.get, self.rearRightMotor.set)



# validated: 2017-11-13 TW 595b1df380f7 edu/wpi/first/wpilibj/Joystick.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import enum
import math
import warnings

import hal
from .interfaces.generichid import GenericHID

__all__ = ["Joystick"]


class Joystick(GenericHID):
    """Handle input from standard Joysticks connected to the Driver Station.
    
    This class handles standard input that comes from the Driver Station. Each
    time a value is requested the most recent value is returned. There is a
    single class instance for each joystick and the mapping of ports to
    hardware buttons depends on the code in the Driver Station.
    """

    kDefaultXAxis = 0
    kDefaultYAxis = 1
    kDefaultZAxis = 2
    kDefaultTwistAxis = 2
    kDefaultThrottleAxis = 3
    kDefaultTriggerButton = 1
    kDefaultTopButton = 2

    class AxisType(enum.IntEnum):
        """Represents an analog axis on a joystick."""
        kX = 0
        kY = 1
        kZ = 2
        kTwist = 3
        kThrottle = 4

    # Note: This is private upstream.
    # This copy is here because of the extra kNumAxes.
    class Axis(enum.IntEnum):
        """Represents an analog axis on a joystick"""
        kX = 0
        kY = 1
        kZ = 2
        kTwist = 3
        kThrottle = 4
        kNumAxes = 5

    class ButtonType(enum.IntEnum):
        """Represents a digital button on the Joystick"""
        kTrigger = 1
        kTop = 2

    # Note: This is private upstream.
    Button = ButtonType

    def __init__(self, port: int) -> None:
        """Construct an instance of a joystick.

        The joystick index is the USB port on the Driver Station.

        This constructor is intended for use by subclasses to configure the
        number of constants for axes and buttons.

        :param port: The port on the Driver Station that the joystick is
            plugged into.
        """
        super().__init__(port)

        self.axes = [0] * self.Axis.kNumAxes
        self.axes[self.Axis.kX] = self.kDefaultXAxis
        self.axes[self.Axis.kY] = self.kDefaultYAxis
        self.axes[self.Axis.kZ] = self.kDefaultZAxis
        self.axes[self.Axis.kTwist] = self.kDefaultTwistAxis
        self.axes[self.Axis.kThrottle] = self.kDefaultThrottleAxis

        hal.report(hal.UsageReporting.kResourceType_Joystick, port)

    def setXChannel(self, channel: int):
        """Set the channel associated with the X axis.

        :param channel: The channel to set the axis to.
        """
        self.axes[self.Axis.kX] = channel

    def setYChannel(self, channel: int):
        """Set the channel associated with the Y axis.

        :param channel: The channel to set the axis to.
        """
        self.axes[self.Axis.kY] = channel

    def setZChannel(self, channel: int):
        """Set the channel associated with the Z axis.

        :param channel: The channel to set the axis to.
        """
        self.axes[self.Axis.kZ] = channel

    def setThrottleChannel(self, channel: int):
        """Set the channel associated with the Throttle axis.

        :param channel: The channel to set the axis to.
        """
        self.axes[self.Axis.kThrottle] = channel

    def setTwistChannel(self, channel: int) -> int:
        """Set the channel associated with the Twist axis.

        :param channel: The channel to set the axis to.
        """
        self.axes[self.Axis.kTwist] = channel

    def setAxisChannel(self, axis, channel):
        """Set the channel associated with a specified axis.

        :param axis: The axis to set the channel for.
        :type  axis: int
        :param channel: The channel to set the axis to.
        :type  channel: int

        .. deprecated:: 2018.0.0
            Use the more specific axis channel setter functions
        """
        warnings.warn("setAxisChannel is deprecated. Use the more specific axis channel setter functions",
                      DeprecationWarning, stacklevel=2)
        self.axes[axis] = channel

    def getXChannel(self) -> int:
        """Get the channel currently associated with the X axis

        :returns: The channel for the axis
        """
        return self.axes[self.Axis.kX]

    def getYChannel(self) -> int:
        """Get the channel currently associated with the Y axis

        :returns: The channel for the axis
        """
        return self.axes[self.Axis.kY]

    def getZChannel(self) -> int:
        """Get the channel currently associated with the Z axis

        :returns: The channel for the axis
        """
        return self.axes[self.Axis.kZ]

    def getThrottleChannel(self) -> int:
        """Get the channel currently associated with the Throttle axis

        :returns: The channel for the axis
        """
        return self.axes[self.Axis.kThrottle]

    def getTwistChannel(self) -> int:
        """Get the channel currently associated with the Twist axis

        :returns: The channel for the axis
        """
        return self.axes[self.Axis.kTwist]

    def getAxisChannel(self, axis: int) -> int:
        """Get the channel currently associated with the specified axis.

        :param axis: The axis to look up the channel for.
        :returns: The channel for the axis.
        
        ..deprecated:: 2018.0.0
            Use the more specific axis channel getter functions
        """
        warnings.warn("getAxisChannel is deprecated. Use the more specific axis channel getter functions",
                      DeprecationWarning, stacklevel=2)
        return self.axes[axis]

    def getX(self, hand=None) -> float:
        """Get the X value of the joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :param hand: Unused
        :returns: The X value of the joystick.
        """
        return self.getRawAxis(self.axes[self.Axis.kX])

    def getY(self, hand=None) -> float:
        """Get the Y value of the joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :param hand: Unused
        :returns: The Y value of the joystick.
        """
        return self.getRawAxis(self.axes[self.Axis.kY])

    def getZ(self, hand=None) -> float:
        """Get the Z position of the HID

        :param hand: Unused
        :returns: the Z position
        """
        return self.getRawAxis(self.axes[self.Axis.kZ])

    def getTwist(self) -> float:
        """Get the twist value of the current joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :returns: The Twist value of the joystick.
        """
        return self.getRawAxis(self.axes[self.AxisType.kTwist])

    def getThrottle(self) -> float:
        """Get the throttle value of the current joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :returns: The Throttle value of the joystick.
        """
        return self.getRawAxis(self.axes[self.AxisType.kThrottle])

    def getAxis(self, axis) -> float:
        """For the current joystick, return the axis determined by the
        argument.

        This is for cases where the joystick axis is returned programmatically,
        otherwise one of the previous functions would be preferable (for
        example :func:`getX`).

        :param axis: The axis to read.
        :type axis: :class:`Joystick.AxisType`
        :returns: The value of the axis.

        ..deprecated: 2018.0.0
            Use the more specific axis getter functions.
        """
        warnings.warn("getAxis is deprecated. Use the more specific axis setter functions",
                      DeprecationWarning, stacklevel=2)

        if axis == self.AxisType.kX:
            return self.getX()
        elif axis == self.AxisType.kY:
            return self.getY()
        elif axis == self.AxisType.kZ:
            return self.getZ()
        elif axis == self.AxisType.kTwist:
            return self.getTwist()
        elif axis == self.AxisType.kThrottle:
            return self.getThrottle()
        else:
            raise ValueError(
                "Invalid axis specified! Must be one of wpilib.Joystick.AxisType, or use getRawAxis instead")

    def getTrigger(self) -> bool:
        """Read the state of the trigger on the joystick.

        Look up which button has been assigned to the trigger and read its
        state.

        :returns: The state of the trigger.
        """
        return self.getRawButton(self.Button.kTrigger)

    def getTriggerPressed(self):
        """Whether the trigger was pressed since the last check

        :returns: Whether the button was pressed since the last check
        """
        return self.getRawButtonPressed(self.Button.kTrigger)

    def getTriggerReleased(self):
        """Whether the trigger was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kTrigger)

    def getTop(self) -> bool:
        """Read the state of the top button on the joystick.

        Look up which button has been assigned to the top and read its state.

        :returns: The state of the top button.
        """
        return self.getRawButton(self.Button.kTop)

    def getTopPressed(self):
        """Whether the trigger was pressed since the last check

        :returns: Whether the button was pressed since the last check
        """
        return self.getRawButtonPressed(self.Button.kTop)

    def getTopReleased(self):
        """Whether the trigger was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kTop)

    def getButton(self, button: ButtonType) -> bool:
        """Get buttons based on an enumerated type.

        The button type will be looked up in the list of buttons and then read.

        :param button: The type of button to read.
        :returns: The state of the button.

        ..deprecated: 2018.0.0
            Use Button enum values instead of ButtonType
        """
        warnings.warn("getButton is deprecated. Use Button enum values instead of ButtonType",
                      DeprecationWarning, stacklevel=2)
        if button == self.ButtonType.kTrigger:
            return self.getTrigger()
        elif button == self.ButtonType.kTop:
            return self.getTop()
        else:
            raise ValueError(
                "Invalid button specified! Must be one of wpilib.Joystick.ButtonType, or use getRawButton instead")

    def getMagnitude(self) -> float:
        """Get the magnitude of the direction vector formed by the joystick's
        current position relative to its origin.

        :returns: The magnitude of the direction vector
        """
        return math.hypot(self.getX(), self.getY())

    def getDirectionRadians(self) -> float:
        """Get the direction of the vector formed by the joystick and its
        origin in radians.

        :returns: The direction of the vector in radians
        """
        return math.atan2(self.getX(), -self.getY())

    def getDirectionDegrees(self) -> float:
        """Get the direction of the vector formed by the joystick and its
        origin in degrees.

        :returns: The direction of the vector in degrees
        """
        return math.degrees(self.getDirectionRadians())



# validated: 2017-09-24 AA e1195e8b9dab edu/wpi/first/wpilibj/PIDOutput.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["PIDOutput"]

class PIDOutput:
    """This interface allows :class:`.PIDController` to write its results to
    its output.
    """

    def pidWrite(self, output):
        """Set the output to the value calculated by :class:`.PIDController`.

        :param output: the value calculated by PIDController
        """
        raise NotImplementedError



# validated: 2018-01-01 EN 40eb6dfc9b83 edu/wpi/first/wpilibj/smartdashboard/SendableBuilderImpl.java
#----------------------------------------------------------------------------
# Copyright (c) 2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------
from networktables import NetworkTables


__all__ = ['SendableBuilder']


class Property:
    def __init__(self, table, key, update, setter):
        self.entry = table.getEntry(key)
        self.key = key
        self.update = update
        self.setter = setter
        self.listener = None

    def createListener(self, entry):
        self.listener = entry.addListener(lambda entry, key, value, param: self.setter(value),
            NetworkTables.NotifyFlags.IMMEDIATE |
            NetworkTables.NotifyFlags.NEW |
            NetworkTables.NotifyFlags.UPDATE)
    
    def startListener(self):
        if self.listener is None and self.setter is not None:
            self.createListener(self.entry)

    def stopListener(self):
        if self.listener is not None:
            self.entry.removeListener(self.listener)
            self.listener = None
        

class SendableBuilder:
    def __init__(self):
        self.table = None
        self._updateTable = None
        self.safeState = None
        self.properties = []

    def setTable(self, table): 
        """
        Set the network table.  Must be called prior to any Add* functions being called.

        :param table: Network table
        :type table: :class:`networktables.networktable.NetworkTable`
        """
        self.table = table

    def getTable(self):
        """
        Get the network table.

        :returns: The network table
        :rtype: :class:`networktables.networktable.NetworkTable`
        """
        return self.table

    def updateTable(self):
        """
        Update the network table values by calling the getters for all properties.
        """
        for prop in self.properties:
            if prop.update is not None:
                prop.update(prop.entry)

        if self._updateTable is not None:
            self._updateTable()

    def startListeners(self):
        """Hook setters for all properties"""
        for prop in self.properties:
            prop.startListener()

    def stopListeners(self):
        """Unhook setters for all properties"""
        for prop in self.properties:
            prop.stopListener()

    def startLiveWindowMode(self):
        """
        Start LiveWindow mode by hooking the setters for all properties. Also calls
        the safeState function if one was provided.
        """
        if self.safeState is not None:
            self.safeState()

        self.startListeners()

    def stopLiveWindowMode(self):
        """
        Stop LiveWindow mode by unhooking the setters for all properties. Also calls
        the safeState function if one was provided.
        """
        self.stopListeners()
        if self.safeState is not None:
            self.safeState()

    def setSmartDashboardType(self, type):
        """
        Set the string representation of the named data type that will be used
        by the smart dashboard for this sendable.
   
        :param type: data type
        :type type: str
        """
        self.table.getEntry(".type").setString(type)

    def setSafeState(self, func):
        """
        Set the function that should be called to set the Sendable into a safe
        state.  This is called when entering and exiting Live Window mode.

        :param func:    function
        """
        self.safeState = func

    def setUpdateTable(self, func):
        """
        Set the function that should be called to update the network table
        for things other than properties.  Note this function is not passed
        the network table object; instead it should use the entry handles
        returned by getEntry().

        :param func:    function
        """
        self._updateTable = func

    def getEntry(self, key):
        """
        Add a property without getters or setters.  This can be used to get
        entry handles for the function called by setUpdateTable().

        :param key:   property name
        :type key: str
        :returns: Network table entry
        :rtype: :class:`networktables.entry.NetworkTableEntry`
        """
        return self.table.getEntry(key)

    def _addProperty(self, key, updater, setter):
        prop = Property(self.table, key, updater, setter)
        self.properties.append(prop)

    def addBooleanProperty(self, key, getter, setter):
        """
        Add a boolean property.
   
        :param key:     property name
        :type getter: str
        :param getter:  getter function (returns current value)
        :type getter: () -> bool
        :param setter:  setter function (sets new value)
        :type setter: (bool) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setBoolean(getter())
        self._addProperty(key, updater, setter)

    def addDoubleProperty(self, key, getter, setter):
        """
        Add a double property.
   
        :param key:     property name
        :type key: str
        :param getter:  getter function (returns current value)
        :type getter: () -> float
        :param setter:  setter function (sets new value)
        :type setter: (float) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setDouble(getter())
        self._addProperty(key, updater, setter)

    def addStringProperty(self, key, getter, setter):
        """
        Add a string property.
   
        :param key:     property name
        :type key: str
        :param getter:  getter function (returns current value)
        :type getter: () -> str
        :param setter:  setter function (sets new value)
        :type setter: (str) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setString(getter())
        self._addProperty(key, updater, setter)

    def addBooleanArrayProperty(self, key, getter, setter):
        """
        Add a boolean array property.
   
        :param key:     property name
        :type key: str
        :param getter:  getter function (returns current value)
        :type getter: () -> bool[]
        :param setter:  setter function (sets new value)
        :type setter: (bool[]) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setBooleanArray(getter())
        self._addProperty(key, updater, setter)

    def addDoubleArrayProperty(self, key, getter, setter):
        """
        Add a double array property.
   
        :param key:     property name
        :type key: str
        :param getter:  getter function (returns current value)
        :type getter: () -> float[]
        :param setter:  setter function (sets new value)
        :type setter: (float[]) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setDoubleArray(getter())
        self._addProperty(key, updater, setter)

    def addStringArrayProperty(self, key, getter, setter):
        """
        Add a string array property.
   
        :param key:     property name
        :type key: str
        :param getter:  getter function (returns current value)
        :type getter: () -> str[]
        :param setter:  setter function (sets new value)
        :type setter: (str[]) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setStringArray(getter())
        self._addProperty(key, updater, setter)

    def addRawProperty(self, key, getter, setter):
        """
        Add a raw property.
   
        :param key:     property name
        :type key: str
        :param getter:  getter function (returns current value)
        :type getter: () -> bytes
        :param setter:  setter function (sets new value)
        :type setter: (bytes) -> Any
        """
        updater = None if getter is None else lambda entry: entry.setRaw(getter())
        self._addProperty(key, updater, setter)



# validated: 2017-12-15 EN f9bece2ffbf7 edu/wpi/first/wpilibj/Jaguar.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .pwmspeedcontroller import PWMSpeedController

__all__ = ["Jaguar"]

class Jaguar(PWMSpeedController):
    """
        Texas Instruments / Vex Robotics Jaguar Speed Controller as a PWM device.
    """

    def __init__(self, channel):
        """Constructor.

        :param channel: The PWM channel that the Jaguar is attached to. 0-9 are on-board, 10-19 are on the MXP port
        """
        super().__init__(channel)
        # Input profile defined by Luminary Micro.
        #
        # Full reverse ranges from 0.671325ms to 0.6972211ms
        # Proportional reverse ranges from 0.6972211ms to 1.4482078ms
        # Neutral ranges from 1.4482078ms to 1.5517922ms
        # Proportional forward ranges from 1.5517922ms to 2.3027789ms
        # Full forward ranges from 2.3027789ms to 2.328675ms
        self.setBounds(2.31, 1.55, 1.507, 1.454, 0.697)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setSpeed(0)
        self.setZeroLatch()
        
        hal.report(hal.UsageReporting.kResourceType_Jaguar,
                   self.getChannel())
        self.setName("Jaguar", self.getChannel())



# validated: 2017-12-02 EN 614093c0c4fc edu/wpi/first/wpilibj/Timer.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import threading

import hal

__all__ = ["Timer"]

class Timer:
    """
        Provides time-related functionality for the robot
        
        .. note:: Prefer to use this module for time functions, instead of
                  the :mod:`time` module in the standard library. This will
                  make it easier for your code to work properly in simulation. 
    """
    
    @staticmethod
    def getFPGATimestamp():
        """Return the system clock time in seconds. Return the time from the
        FPGA hardware clock in seconds since the FPGA started.

        :returns: Robot running time in seconds.
        :rtype: float
        """
        return hal.getFPGATime() / 1000000.0

    @staticmethod
    def getMatchTime():
        """Return the approximate match time.
        The FMS does not currently send the official match time to the robots,
        but does send an approximate match time. The value will count down
        the time remaining in the current period (auto or teleop). 
        
        .. warning::
            This is not an official time (so it cannot be used to dispute ref
            calls or guarantee that a function will trigger before the match ends).

        The Practice Match function of the DS approximates the behavior seen
        on the field.

        :returns: Time remaining in current match period (auto or teleop) in seconds
        :rtype: float
        """
        from .driverstation import DriverStation
        return DriverStation.getInstance().getMatchTime()

    @staticmethod
    def delay(seconds):
        """Pause the thread for a specified time. Pause the execution of the
        thread for a specified period of time given in seconds. Motors will
        continue to run at their last assigned values, and sensors will
        continue to update. Only the thread containing the wait will pause
        until the wait time is expired.

        :param seconds: Length of time to pause
        :type seconds: float
        
        .. warning:: If you're tempted to use this function for autonomous
                     mode to time transitions between actions, don't do it!
                     
                     Delaying the main robot thread for more than a few
                     milliseconds is generally discouraged, and will cause
                     problems and possibly leave the robot unresponsive.
                     
        """
        hal.sleep(seconds)

    def __init__(self):
        self.mutex = threading.RLock()
        self.startTime = self.getMsClock()
        self.accumulatedTime = 0.0
        self.running = False

    def getMsClock(self):
        """
            :returns: the system clock time in milliseconds.
            :rtype: int
        """
        return hal.getFPGATime() / 1000

    def get(self):
        """Get the current time from the timer. If the clock is running it is
        derived from the current system clock the start time stored in the
        timer class. If the clock is not running, then return the time when
        it was last stopped.

        :returns: Current time value for this timer in seconds
        :rtype: float
        """
        with self.mutex:
            if self.running:
                return ((self.getMsClock() - self.startTime) + self.accumulatedTime) / 1000.0
            else:
                return self.accumulatedTime

    def reset(self):
        """Reset the timer by setting the time to 0.
        Make the timer startTime the current time so new requests will be
        relative now.
        """
        with self.mutex:
            self.accumulatedTime = 0.0
            self.startTime = self.getMsClock()

    def start(self):
        """Start the timer running.
        Just set the running flag to true indicating that all time requests
        should be relative to the system clock.
        """
        with self.mutex:
            self.startTime = self.getMsClock()
            self.running = True

    def stop(self):
        """Stop the timer.
        This computes the time as of now and clears the running flag, causing
        all subsequent time requests to be read from the accumulated time
        rather than looking at the system clock.
        """
        with self.mutex:
            temp = self.get()
            self.accumulatedTime = temp
            self.running = False
            
    def hasPeriodPassed(self, period):
        """Check if the period specified has passed and if it has, advance the start
        time by that period. This is useful to decide if it's time to do periodic
        work without drifting later by the time it took to get around to checking.
 
        :param period: The period to check for (in seconds).
        :returns: If the period has passed.
        :rtype: bool
        """
        
        with self.mutex:
            if self.get() > period:
                # Advance the start time by the period
                # Don't set it to the current time... we want to avoid drift
                self.startTime += (period * 1000)
                return True
            
            return False



# validated: 2018-01-01 EN 40eb6dfc9b83 edu/wpi/first/wpilibj/IterativeRobotBase.java
# ----------------------------------------------------------------------------
# Copyright (c) 2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import enum
import logging

import hal
from .livewindow import LiveWindow
from .smartdashboard import SmartDashboard
from .robotbase import RobotBase

__all__ = ["IterativeRobotBase"]


class IterativeRobotBase(RobotBase):
    """IterativeRobotBase implements a specific type of robot program framework, extending the RobotBase
    class.

    The IterativeRobotBase class does not implement startCompetition(), so it should not be used
    by teams directly.

    This class provides the following functions which are called by the main loop,
    startCompetition(), at the appropriate times:

    robotInit() -- provide for initialization at robot power-on

    init() functions -- each of the following functions is called once when the
    appropriate mode is entered:
    - disabledInit()   -- called only when first disabled
    - autonomousInit() -- called each and every time autonomous is entered from
    another mode
    - teleopInit()     -- called each and every time teleop is entered from
    another mode
    - testInit()       -- called each and every time test is entered from
    another mode

    periodic() functions -- each of these functions is called on an interval:
    - robotPeriodic()
    - disabledPeriodic()
    - autonomousPeriodic()
    - teleopPeriodic()
    - testPeriodic()
    """

    class Mode(enum.Enum):
        kNone = 0
        kDisabled = 1
        kAutonomous = 2
        kTeleop = 3
        kTest = 4

    #: A python logging object that you can use to send messages to the log. It
    #: is recommended to use this instead of print statements.
    logger = logging.getLogger("robot")

    # ----------- Overridable initialization code -----------------

    def __init__(self):
        self.last_mode = self.Mode.kNone
        super().__init__()

    def robotInit(self):
        """Robot-wide initialization code should go here.

        Users should override this method for default Robot-wide initialization
        which will be called when the robot is first powered on.  It will be
        called exactly 1 time.

        .. note:: It is simpler to override this function instead of defining
                  a constructor for your robot class
        """
        self.logger.info("Default IterativeRobot.robotInit() method... Overload me!")

    def disabledInit(self):
        """Initialization code for disabled mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters disabled mode.
        """
        self.logger.info("Default IterativeRobot.disabledInit() method... Overload me!")

    def autonomousInit(self):
        """Initialization code for autonomous mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters autonomous mode.
        """
        self.logger.info("Default IterativeRobot.autonomousInit() method... Overload me!")

    def teleopInit(self):
        """Initialization code for teleop mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters teleop mode.
        """
        self.logger.info("Default IterativeRobot.teleopInit() method... Overload me!")

    def testInit(self):
        """Initialization code for test mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters test mode.
        """
        self.logger.info("Default IterativeRobot.testInit() method... Overload me!")

    # ----------- Overridable periodic code -----------------

    def robotPeriodic(self):
        """Periodic code for all robot modes should go here."""
        func = self.robotPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.robotPeriodic() method... Overload me!")
            func.firstRun = False

    def disabledPeriodic(self):
        """Periodic code for disabled mode should go here."""
        func = self.disabledPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.disabledPeriodic() method... Overload me!")
            func.firstRun = False

    def autonomousPeriodic(self):
        """Periodic code for autonomous mode should go here."""
        func = self.autonomousPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.autonomousPeriodic() method... Overload me!")
            func.firstRun = False

    def teleopPeriodic(self):
        """Periodic code for teleop mode should go here."""
        func = self.teleopPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warning("Default IterativeRobot.teleopPeriodic() method... Overload me!")
            func.firstRun = False

    def testPeriodic(self):
        """Periodic code for test mode should go here."""
        func = self.testPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.testPeriodic() method... Overload me!")
            func.firstRun = False

    def loopFunc(self):
        """Call the appropriate function depending upon the current robot mode"""

        if self.isDisabled():
            if self.last_mode is not self.Mode.kDisabled:
                LiveWindow.setEnabled(False)
                self.disabledInit()
                self.last_mode = self.Mode.kDisabled
            hal.observeUserProgramDisabled()
            self.disabledPeriodic()
        elif self.isAutonomous():
            if self.last_mode is not self.Mode.kAutonomous:
                LiveWindow.setEnabled(False)
                self.autonomousInit()
                self.last_mode = self.Mode.kAutonomous
            hal.observeUserProgramAutonomous()
            self.autonomousPeriodic()
        elif self.isOperatorControl():
            if self.last_mode is not self.Mode.kTeleop:
                LiveWindow.setEnabled(False)
                self.teleopInit()
                self.last_mode = self.Mode.kTeleop
            hal.observeUserProgramTeleop()
            self.teleopPeriodic()
        else:
            if self.last_mode is not self.Mode.kTest:
                LiveWindow.setEnabled(False)
                self.testInit()
                self.last_mode = self.Mode.kTest
            hal.observeUserProgramTest()
            self.testPeriodic()
        self.robotPeriodic()
        SmartDashboard.updateValues()
        LiveWindow.updateValues()



# validated: 2017-11-13 TW 21585f70a88e edu/wpi/first/wpilibj/GamepadBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import warnings

from .generichid import GenericHID

__all__ = ["GamepadBase"]


class GamepadBase(GenericHID):
    """
    GamepadBase Interface.

    .. deprecated: 2018.0.0
       Inherit directly from :class:`.GenericHID` instead.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        warnings.warn("GamepadBase is deprecated, inherit directly from GenericHID instead",
                      DeprecationWarning, stacklevel=2)

    def getRawAxis(self, axis):
        raise NotImplementedError

    def getBumper(self, hand):
        """Is the bumper pressed.

        :param hand: which hand
        :returns: true if the bumper is pressed
        """
        raise NotImplementedError

    def getStickButton(self, hand=None):
        raise NotImplementedError

    def getRawButton(self, button):
        raise NotImplementedError

    def getPOV(self, pov=0):
        raise NotImplementedError

    def getPOVCount(self):
        raise NotImplementedError

    def getType(self):
        raise NotImplementedError

    def getName(self):
        raise NotImplementedError

    def setOutput(self, outputNumber, value):
        raise NotImplementedError

    def setOutputs(self, value):
        raise NotImplementedError

    def setRumble(self, type, value):
        raise NotImplementedError



# validated: 2017-12-23 EN f9bece2ffbf7 edu/wpi/first/wpilibj/DigitalInput.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal

from .digitalsource import DigitalSource

__all__ = ["DigitalInput"]


class DigitalInput(DigitalSource):
    """Reads a digital input.
    
    This class will read digital inputs and return the current value on the
    channel. Other devices such as encoders, gear tooth sensors, etc. that
    are implemented elsewhere will automatically allocate digital inputs
    and outputs as required. This class is only for devices like switches
    etc. that aren't implemented anywhere else.
    """

    def __init__(self, channel):
        """Create an instance of a Digital Input class. Creates a digital
        input given a channel.

        :param channel: the DIO channel for the digital input. 0-9 are on-board, 10-25 are on the MXP
        :type  channel: int
        """

        super().__init__()
        self.checkDigitalChannel(channel)
        self.channel = channel

        self.handle = hal.initializeDIOPort(hal.getPort(channel), True)

        hal.report(hal.UsageReporting.kResourceType_DigitalInput,
                   channel)
        self.setName("DigitalInput", channel)

    def free(self):
        super().free()
        if self.interrupt:
            self.cancelInterrupts()

        hal.freeDIOPort(self.handle)

    def get(self):
        """Get the value from a digital input channel. Retrieve the value of
        a single digital input channel from the FPGA.

        :returns: the state of the digital input
        :rtype: bool
        """
        return hal.getDIO(self.handle)

    def getChannel(self):
        """Get the channel of the digital input.

        :returns: The GPIO channel number that this object represents.
        :rtype: int
        """
        return self.channel

    def getAnalogTriggerTypeForRouting(self):
        """Get the analog trigger type.

        :returns: false
        :rtype: int
        """
        return 0

    def isAnalogTrigger(self):
        """Is this an analog trigger.

        :returns: true if this is an analog trigger
        :rtype: bool
        """
        return False

    def getPortHandleForRouting(self):
        """Get the HAL Port Handle.

        :return: The HAL Handle to the specified source
        """
        return self.handle

    def initSendable(self, builder):
        builder.setSmartDashboardType("Digital Input")
        builder.addBooleanProperty("Value", self.get, None)



# validated: 2018-01-19 DS a3e5378d1468 edu/wpi/first/wpilibj/SPI.java

import ctypes
import enum
import hal
import struct
import threading
import warnings
import weakref

from .accumulatorresult import AccumulatorResult
from .digitalsource import DigitalSource
from .notifier import Notifier

__all__ = ["SPI"]

def _freeSPI(port):
    hal.closeSPI(port)

class SPI:
    """Represents a SPI bus port

    Example usage::

        spi = wpilib.SPI(wpilib.SPI.Port.kOnboardCS0)

        # Write bytes 'text', and receive something
        data = spi.transaction(b'text')

    """

    class Port(enum.IntEnum):
        kOnboardCS0 = 0
        kOnboardCS1 = 1
        kOnboardCS2 = 2
        kOnboardCS3 = 3
        kMXP = 4

    devices = 0

    @staticmethod
    def _reset():
        SPI.devices = 0

    def __init__(self, port, simPort=None):
        """Constructor

        :param port: the physical SPI port
        :type port: :class:`.SPI.Port`
        :param simPort: This must be an object that implements all of
                        the spi* functions from hal_impl that you use.
                        See ``test_spi.py`` for an example.
        """

        port = self.Port(port)

        # python-specific
        if hal.HALIsSimulation():
            if simPort is None:
                # If you want more functionality, implement your own mock
                from hal_impl.spi_helpers import SPISimBase
                simPort = SPISimBase()

                msg = "Using stub simulator for SPI port %s" % port
                warnings.warn(msg)

            # Just check for basic functionality
            assert hasattr(simPort, 'initializeSPI')
            assert hasattr(simPort, 'closeSPI')

            self._port = (simPort, port)
        else:
            self._port = port

        # python-specific: these are bools instead of integers
        self.bitOrder = False
        self.clockPolarity = False
        self.dataOnTrailing = False

        self.accum = None

        hal.initializeSPI(self._port)
        self.__finalizer = weakref.finalize(self, _freeSPI, self._port)

        SPI.devices += 1
        hal.report(hal.UsageReporting.kResourceType_SPI, SPI.devices)

    @property
    def port(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use SPI after free() has been called")
        return self._port

    def free(self):
        if self.accum is not None:
            self.accum.free()
            self.accum = None
        self.__finalizer()

    def setClockRate(self, hz: int) -> None:
        """Configure the rate of the generated clock signal. The default value is 500,000 Hz. The maximum
        value is 4,000,000 Hz.

        :param hz: The clock rate in Hertz.
        """
        hal.setSPISpeed(self.port, hz)

    def setMSBFirst(self) -> None:
        """Configure the order that bits are sent and received on the wire to be most significant bit
        first.
        """
        self.bitOrder = True
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setLSBFirst(self) -> None:
        """Configure the order that bits are sent and received on the wire to be least significant bit
        first.
        """
        self.bitOrder = False
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setClockActiveLow(self) -> None:
        """Configure the clock output line to be active low. This is sometimes called clock polarity high
        or clock idle high.
        """
        self.clockPolarity = True
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setClockActiveHigh(self) -> None:
        """Configure the clock output line to be active high. This is sometimes called clock polarity low
        or clock idle low.
        """
        self.clockPolarity = False
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setSampleDataOnFalling(self) -> None:
        """Configure that the data is stable on the falling edge and the data changes on the rising edge."""
        self.dataOnTrailing = True
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setSampleDataOnRising(self) -> None:
        """Configure that the data is stable on the rising edge and the data changes on the falling edge."""
        self.dataOnTrailing = False
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setChipSelectActiveHigh(self) -> None:
        """Configure the chip select line to be active high."""
        hal.setSPIChipSelectActiveHigh(self.port)

    def setChipSelectActiveLow(self) -> None:
        """Configure the chip select line to be active low."""
        hal.setSPIChipSelectActiveLow(self.port)

    def write(self, dataToSend: bytes) -> int:
        """Write data to the slave device.  Blocks until there is space in the
        output FIFO.

        If not running in output only mode, also saves the data received
        on the MISO input during the transfer into the receive FIFO.

        :param dataToSend: Data to send
        :type dataToSend: iterable of bytes

        :returns: Number of bytes written

        Usage::

            # send byte string
            writeCount = spi.write(b'stuff')

            # send list of integers
            writeCount = spi.write([0x01, 0x02])
        """
        return hal.writeSPI(self.port, dataToSend)

    def read(self, initiate: bool, size: int) -> bytes:
        """Read a word from the receive FIFO.

        Waits for the current transfer to complete if the receive FIFO is
        empty.

        If the receive FIFO is empty, there is no active transfer, and
        initiate is False, errors.

        :param initiate: If True, this function pushes "0" into the
            transmit buffer and initiates a transfer.  If False, this function
            assumes that data is already in the receive FIFO from a previous
            write.
        :param size: Number of bytes to read.

        :returns: received data bytes
        """
        if initiate:
            return hal.transactionSPI(self.port, [0]*size)
        else:
            return hal.readSPI(self.port, size)

    def transaction(self, dataToSend: bytes) -> bytes:
        """Perform a simultaneous read/write transaction with the device

        :param dataToSend: The data to be written out to the device
        :type dataToSend: iterable of bytes

        :returns: data received from the device

        Usage::

            # send byte string
            data = spi.transaction(b'stuff')

            # send list of integers
            data = spi.transaction([0x01, 0x02])
        """
        return hal.transactionSPI(self.port, dataToSend)

    def initAuto(self, bufferSize: int) -> None:
        """Initialize automatic SPI transfer engine.

        Only a single engine is available, and use of it blocks use of all other
        chip select usage on the same physical SPI port while it is running.

        :param bufferSize: buffer size in bytes
        """
        hal.initSPIAuto(self.port, bufferSize)

    def freeAuto(self) -> None:
        """Frees the automatic SPI transfer engine."""
        hal.freeSPIAuto(self.port)

    def setAutoTransmitData(self, dataToSend: bytes, zeroSize: int) -> None:
        """Set the data to be transmitted by the engine.

        Up to 16 bytes are configurable, and may be followed by up to 127 zero
        bytes.

        :param dataToSend: data to send (maximum 16 bytes)
        :param zeroSize: number of zeros to send after the data
        """
        hal.setSPIAutoTransmitData(self.port, dataToSend, zeroSize)

    def startAutoRate(self, period: float) -> None:
        """Start running the automatic SPI transfer engine at a periodic rate.

        :meth:`.initAuto` and :meth:`.setAutoTransmitData` must
        be called before calling this function.

        :param period: period between transfers, in seconds (us resolution)
        """
        hal.startSPIAutoRate(self.port, period)

    def startAutoTrigger(self, source: DigitalSource, rising: bool, falling: bool) -> None:
        """Start running the automatic SPI transfer engine when a trigger occurs.

        :meth:`.initAuto` and :meth:`.setAutoTransmitData` must
        be called before calling this function.

        :param source: digital source for the trigger (may be an analog trigger)
        :param rising: trigger on the rising edge
        :param falling: trigger on the falling edge
        """
        hal.startSPIAutoTrigger(self.port, source.getPortHandleForRouting(),
                                source.getAnalogTriggerTypeForRouting(),
                                rising, falling)

    def stopAuto(self) -> None:
        """Stop running the automatic SPI transfer engine."""
        hal.stopSPIAuto(self.port)

    def forceAutoRead(self) -> None:
        """Force the engine to make a single transfer."""
        hal.forceSPIAutoRead(self.port)

    def readAutoReceivedData(self, buffer, numToRead: int, timeout: float) -> (int, bytes):
        """Read data that has been transferred by the automatic SPI transfer engine.

        Transfers may be made a byte at a time, so it's necessary for the caller
        to handle cases where an entire transfer has not been completed.

        Blocks until numToRead bytes have been read or timeout expires.
        May be called with numToRead=0 to retrieve how many bytes are available.

        :param buffer: A ctypes c_uint8 buffer to read the data into
        :param numToRead: number of bytes to read
        :param timeout: timeout in seconds (ms resolution)
        :returns: Number of bytes remaining to be read
        """
        if len(buffer) < numToRead:
            raise ValueError("buffer is too small, must be at least %s" % numToRead)
        return hal.readSPIAutoReceivedData(self.port, buffer, numToRead, timeout)

    def getAutoDroppedCount(self) -> int:
        """Get the number of bytes dropped by the automatic SPI transfer engine due
        to the receive buffer being full.

        :returns: Number of bytes dropped
        """
        return hal.getSPIAutoDroppedCount(self.port)

    class _Accumulator:

        kAccumulateDepth = 2048

        def __init__(self, port: int, xferSize: int, validMask: int,
                           validValue: int, dataShift: int, dataSize: int,
                           isSigned: bool, bigEndian: bool):

            # python-specific: for efficiency purposes, we only support xferSize of
            #                  2, 4, or 8... if you need to do a different size then
            #                  file a bug and we'll adjust it
            efmt = '>' if bigEndian else '<'
            self._xferSize = xferSize                   # SPI transfer size, in bytes
            if self._xferSize == 2:
                self._struct = struct.Struct('%sH' % efmt)
            elif self._xferSize == 4:
                self._struct = struct.Struct('%sI' % efmt)
            elif self._xferSize == 8:
                self._struct = struct.Struct('%sQ' % efmt)
            else:
                raise ValueError("SPI Accumulator only supports xferSize of 2/4/8")

            self._mutex = threading.RLock()
            self._notifier = Notifier(self._update)
            self._buf = (ctypes.c_uint8 * (xferSize * self.kAccumulateDepth))()

            self._validMask = validMask
            self._validValue = validValue
            self._dataShift = dataShift                 # data field shift right amount, in bits
            self._dataMax = 1 << dataSize               # one more than max data value
            self._dataMsbMask = 1 << (dataSize - 1)     # data field MSB mask (for signed)
            self._isSigned = isSigned                   # is data field signed?
            self._bigEndian = bigEndian                 # is response big endian?
            self._port = port

            self._value = 0
            self._count = 0
            self._lastValue = 0

            self._center = 0
            self._deadband = 0

        def free(self):
            self._notifier.stop()

        def _update(self):
            with self._mutex:
                done = False
                while not done:
                    done = True

                    # get amount of data available
                    numToRead = hal.readSPIAutoReceivedData(self._port, self._buf, 0, 0)

                    # only get whole responses
                    numToRead -= numToRead % self._xferSize
                    if numToRead > self._xferSize * self.kAccumulateDepth:
                        numToRead = self._xferSize * self.kAccumulateDepth
                        done = False

                    if numToRead == 0:
                        return  # no samples

                    # read buffered data
                    hal.readSPIAutoReceivedData(self._port, self._buf, numToRead, 0)

                    # loop over all responses
                    off = 0
                    while True:
                        if off > numToRead:
                            break

                        # convert from bytes
                        resp = self._struct.unpack_from(self._buf, off)[0]

                        # process response
                        if (resp & self._validMask) == self._validValue:

                            # valid sensor data extract data field
                            data = resp >> self._dataShift
                            data &= self._dataMax - 1

                            # 2s complement conversion if signed MSB is set
                            if (self._isSigned and (data & self._dataMsbMask) != 0):
                                data -= self._dataMax

                            # center offset
                            data -= self._center

                            # only accumulate if outside deadband
                            if (data < -self._deadband or data > self._deadband):
                                self._value += data

                            self._count += 1
                            self._lastValue = data
                        else:
                            # no data from the sensor just clear the last value
                            self._lastValue = 0

                        off += self._xferSize


    def initAccumulator(self, period: float, cmd: int, xferSize: int,
                              validMask: int, validValue: int,
                              dataShift: int, dataSize: int,
                              isSigned: bool, bigEndian: bool) -> None:
        """Initialize the accumulator.

        :param period: Time between reads
        :param cmd: SPI command to send to request data
        :param xferSize: SPI transfer size, in bytes
        :param validMask: Mask to apply to received data for validity checking
        :param validValue: After validMask is applied, required matching value for validity checking
        :param dataShift: Bit shift to apply to received data to get actual data value
        :param dataSize: Size (in bits) of data field
        :param isSigned: Is data field signed?
        :param bigEndian: Is device big endian?
        """
        self.initAuto(xferSize * 2048)

        if bigEndian:
            cmdBytes = cmd.to_bytes(4, 'big')
        else:
            cmdBytes = cmd.to_bytes(4, 'little')

        self.setAutoTransmitData(cmdBytes, xferSize - 4)
        self.startAutoRate(period)

        self.accum = self._Accumulator(self.port, xferSize, validMask, validValue,
                                      dataShift, dataSize, isSigned, bigEndian)
        self.accum._notifier.startPeriodic(period * 1024)

    def freeAccumulator(self) -> None:
        """Frees the accumulator."""
        if self.accum:
            self.accum.free()
            self.accum = None
        self.freeAuto()

    def resetAccumulator(self) -> None:
        """Resets the accumulator to zero."""
        if not self.accum:
            return
        with self.accum._mutex:
            self.accum._value = 0
            self.accum._count = 0
            self.accum._lastValue = 0

    def setAccumulatorCenter(self, center: int) -> None:
        """Set the center value of the accumulator.

        The center value is subtracted from each value before it is added to the accumulator. This
        is used for the center value of devices like gyros and accelerometers to make integration work
        and to take the device offset into account when integrating.
        """
        if not self.accum:
            return
        with self.accum._mutex:
            self.accum._center = center

    def setAccumulatorDeadband(self, deadband: int) -> None:
        """Set the accumulator's deadband."""
        if not self.accum:
            return
        with self.accum._mutex:
            self.accum._deadband = deadband

    def getAccumulatorLastValue(self) -> int:
        """Read the last value read by the accumulator engine."""
        if not self.accum:
            return 0
        with self.accum._mutex:
            self.accum._update()
            return self.accum._lastValue

    def getAccumulatorValue(self) -> int:
        """Read the accumulated value.

        :returns: The 64-bit value accumulated since the last Reset().
        """
        if not self.accum:
            return 0
        with self.accum._mutex:
            self.accum._update()
            return self.accum._value

    def getAccumulatorCount(self) -> int:
        """Read the number of accumulated values.

        Read the count of the accumulated values since the accumulator was last Reset().

        :returns: The number of times samples from the channel were accumulated.
        """
        if not self.accum:
            return 0
        with self.accum._mutex:
            self.accum._update()
            return self.accum._count

    def getAccumulatorAverage(self) -> float:
        """Read the average of the accumulated value.

        :returns: The accumulated average value (value / count).
        """
        if not self.accum:
            return 0
        with self.accum._mutex:
            self.accum._update()
            if self.accum._count == 0:
                return 0.0
            else:
                return float(self.accum._value) / self.accum._count

    def getAccumulatorOutput(self) -> AccumulatorResult:
        """Read the accumulated value and the number of accumulated values atomically.

        This function reads the value and count atomically.
        This can be used for averaging.

        :returns: tuple of (value, count)
        """
        if self.accum is None:
            return AccumulatorResult(0, 0)

        with self.accum._mutex:
            self.accum._update()
            return AccumulatorResult(self.accum._value, self.accum._count)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from collections import namedtuple

__all__ = ['AccumulatorResult']

AccumulatorResult = namedtuple('AccumulatorResult', ['value', 'count'])

AccumulatorResult.__doc__ = 'Structure for holding the values stored in an accumulator.'
AccumulatorResult.value.__doc__ = 'The total value accumulated.'
AccumulatorResult.count.__doc__ = 'The number of sample value was accumulated over.'



'''
    This is the core of WPILib.
'''

from .accumulatorresult import *
from .adxl345_i2c import *
from .adxl345_spi import *
from .adxl362 import *
from .adxrs450_gyro import *
from .analogaccelerometer import *
from .analoginput import *
from .analoggyro import *
from .analogoutput import *
from .analogpotentiometer import *
from .analogtrigger import *
from .analogtriggeroutput import *
from .builtinaccelerometer import *
from .cameraserver import *
from .canjaguar import *
from .cantalon import *
from .compressor import *
from .controllerpower import *
from .counter import *
from .digitalglitchfilter import *
from .digitalinput import *
from .digitaloutput import *
from .digitalsource import *
from .dmc60 import *
from .doublesolenoid import *
from .driverstation import *
from .encoder import *
from .filter import *
from .geartooth import *
from .gyrobase import *
from .i2c import *
from .interruptablesensorbase import *
from .iterativerobot import *
from .iterativerobotbase import *
from .jaguar import *
from .joystick import *
from .lineardigitalfilter import *
from .livewindow import *
from .livewindowsendable import *
from .motorsafety import *
from .nidecbrushless import *
from .notifier import *
from .pidcontroller import *
from .powerdistributionpanel import *
from .preferences import *
from .pwm import *
from .pwmspeedcontroller import *
from .pwmtalonsrx import *
from .pwmvictorspx import *
from .relay import *
from .resource import *
from .robotbase import *
from .robotcontroller import *
from .robotdrive import *
from .robotstate import *
from .safepwm import *
from .samplerobot import *
from .sd540 import *
from .sendable import *
from .sendablebase import *
from .sendablebuilder import *
from .sendablechooser import *
from .sensorbase import *
from .serialport import *
from .servo import *
from .smartdashboard import *
from .solenoidbase import *
from .solenoid import *
from .spark import *
from .speedcontrollergroup import *
from .spi import *
from .talon import *
from .timedrobot import *
from .timer import *
from .ultrasonic import *
from .utility import *
from .victor import *
from .victorsp import *
from .xboxcontroller import *

from ._impl.main import run

try:
    from .version import __version__
except ImportError:
    __version__ = 'master'



# validated: 2017-11-09 TW ef3267833fc3 edu/wpi/first/wpilibj/IterativeRobot.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal
import logging

from .iterativerobotbase import IterativeRobotBase

__all__ = ["IterativeRobot"]


class IterativeRobot(IterativeRobotBase):
    """IterativeRobot implements the IterativeRobotBase robot program framework.

    The IterativeRobot class is intended to be subclassed by a user creating a robot program.

    periodic() functions from the base class are called each time a new packet is received from
    the driver station.
    """

    #: A python logging object that you can use to send messages to the log. It
    #: is recommended to use this instead of print statements.
    logger = logging.getLogger("robot")

    def __init__(self):
        """Constructor for RobotIterativeBase.

        The constructor initializes the instance variables for the robot to
        indicate the status of initialization for disabled, autonomous, and
        teleop code.

        .. warning:: If you override ``__init__`` in your robot class, you must call
                     the base class constructor. This must be used to ensure that
                     the communications code starts.
        """
        super().__init__()

        hal.report(hal.UsageReporting.kResourceType_Framework,
                   hal.UsageReporting.kFramework_Iterative)

    def startCompetition(self):
        """Provide an alternate "main loop" via startCompetition()."""

        self.robotInit()

        # Tell the DS that the robot is ready to be enabled
        hal.observeUserProgramStarting()

        # loop forever, calling the appropriate mode-dependent function
        while True:
            # Wait for new data to arrive
            self.ds.waitForData()
            # Call the appropriate function depending upon the current robot mode
            self.loopFunc()



# notrack

import hal
import threading

import logging
logger = logging.getLogger('wpilib.cs')

__all__ = ['CameraServer']

class CameraServer:
    '''
        Provides a way to launch an out of process cscore-based camera
        service instance, for streaming or for image processing.
        
        .. note:: This does not correspond directly to the wpilib
                  CameraServer object; that can be found as
                  :class:`cscore.CameraServer`. However, you should
                  not use cscore directly from your robot code, see
                  the documentation for details             
    '''
    
    _alive = False
    _launched = False
    
    @classmethod
    def is_alive(cls):
        ''':returns: True if the CameraServer is still alive'''
        return cls._alive
    
    @classmethod
    def launch(cls, vision_py=None):
        '''
            Launches the CameraServer process in autocapture mode or
            using a user-specified python script
        
            :param vision_py: If specified, this is the relative path to
                              a filename with a function in it
                              
            Example usage::
            
                wpilib.CameraServer.launch("vision.py:main")
            
            .. warning:: You must have robotpy-cscore installed, or this
                         function will fail without returning an error
                         (you will see an error in the console).
                         
        '''
        
        if cls._launched:
            return
        
        cls._launched = True
        
        if hal.isSimulation():
            logger.info("Would launch CameraServer with vision_py=%s", vision_py)
            cls._alive = True
        else:
            logger.info("Launching CameraServer process")
            
            # Launch the cscore launcher in a separate process
            
            import subprocess
            import sys
            
            args = [
                sys.executable,
                '-m', 'cscore'
            ]
            
            if vision_py:
                if not vision_py.startswith('/'):
                    vision_py = '/home/lvuser/py/' + vision_py
                args.append(vision_py)
            
            # We open a pipe to it so that when this process exits, it dies
            proc = subprocess.Popen(args, close_fds=True, stdin=subprocess.PIPE, cwd='/home/lvuser/py')
            th = threading.Thread(target=cls._monitor_child, args=(proc,))
            th.daemon = True
            th.start()
           
    @classmethod 
    def _monitor_child(cls, proc):
        proc.wait()
        logger.warning("CameraServer process exited with exitcode %s", proc.returncode)
        cls._alive = False
                


# validated: 2017-12-16 EN f9bece2ffbf7 edu/wpi/first/wpilibj/command/Scheduler.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from ..sendablebase import SendableBase

import collections
import warnings

__all__ = ["Scheduler"]

class Scheduler(SendableBase):
    """The Scheduler is a singleton which holds the top-level running commands.
    It is in charge of both calling the command's run() method and to make
    sure that there are no two commands with conflicting requirements running.

    It is fine if teams wish to take control of the Scheduler themselves, all
    that needs to be done is to call Scheduler.getInstance().run() often to
    have Commands function correctly. However, this is already done for you
    if you use the CommandBased Robot template.

    .. seealso:: :class:`.Command`
    """
    
    @staticmethod
    def _reset():
        try:
            del Scheduler.instance
        except:
            pass

    @staticmethod
    def getInstance():
        """Returns the Scheduler, creating it if one does not exist.

        :returns: the Scheduler
        """
        if not hasattr(Scheduler, "instance"):
            Scheduler.instance = Scheduler()
        return Scheduler.instance

    def __init__(self):
        """Instantiates a Scheduler.
        """
        super().__init__()
        hal.report(hal.UsageReporting.kResourceType_Command,
                   hal.UsageReporting.kCommand_Scheduler)
        self.setName("Scheduler")

        # Active Commands
        self.commandTable = collections.OrderedDict()
        # The set of all Subsystems
        self.subsystems = set()
        # Whether or not we are currently adding a command
        self.adding = False
        # Whether or not we are currently disabled
        self.disabled = False
        # A list of all Commands which need to be added
        self.additions = []
        # A list of all Buttons. It is created lazily.
        self.buttons = []
        self.runningCommandsChanged = False

        self.namesEntry = None
        self.idsEntry = None
        self.cancelEntry = None

    def add(self, command):
        """Adds the command to the Scheduler. This will not add the
        :class:`.Command` immediately, but will instead wait for the proper time in
        the :meth:`run` loop before doing so. The command returns immediately
        and does nothing if given null.

        Adding a :class:`.Command` to the :class:`.Scheduler` involves the
        Scheduler removing any Command which has shared requirements.

        :param command: the command to add
        """
        if command is not None:
            self.additions.append(command)

    def addButton(self, button):
        """Adds a button to the Scheduler. The Scheduler will poll
        the button during its :meth:`run`.

        :param button: the button to add
        """
        self.buttons.append(button)

    def _add(self, command):
        """Adds a command immediately to the Scheduler. This should only be
        called in the :meth:`run` loop. Any command with conflicting
        requirements will be removed, unless it is uninterruptable. Giving
        None does nothing.

        :param command: the :class:`.Command` to add
        """
        if command is None:
            return

        # Check to make sure no adding during adding
        if self.adding:
            warnings.warn("Can not start command from cancel method.  Ignoring: %s" % command, RuntimeWarning)
            return

        # Only add if not already in
        if command not in self.commandTable:
            # Check that the requirements can be had
            for lock in command.getRequirements():
                if (lock.getCurrentCommand() is not None and
                    not lock.getCurrentCommand().isInterruptible()):
                    return

            # Give it the requirements
            self.adding = True
            for lock in command.getRequirements():
                if lock.getCurrentCommand() is not None:
                    lock.getCurrentCommand().cancel()
                    self.remove(lock.getCurrentCommand())
                lock.setCurrentCommand(command)
            self.adding = False

            # Add it to the list
            self.commandTable[command] = 1

            self.runningCommandsChanged = True

            command.startRunning()

    def run(self):
        """Runs a single iteration of the loop. This method should be called
        often in order to have a functioning Command system. The loop has five
        stages:

        - Poll the Buttons
        - Execute/Remove the Commands
        - Send values to SmartDashboard
        - Add Commands
        - Add Defaults
        """

        self.runningCommandsChanged = False

        if self.disabled:
            return # Don't run when disabled

        # Get button input (going backwards preserves button priority)
        for button in reversed(self.buttons):
            button()

        # Call every subsystem's periodic method
        for subsystem in self.subsystems:
            subsystem.periodic()

        # Loop through the commands
        for command in list(self.commandTable):
            if not command.run():
                self.remove(command)
                self.runningCommandsChanged = True

        # Add the new things
        for command in self.additions:
            self._add(command)
        self.additions.clear()

        # Add in the defaults
        for lock in self.subsystems:
            if lock.getCurrentCommand() is None:
                self._add(lock.getDefaultCommand())
            lock.confirmCommand()

    def registerSubsystem(self, system):
        """Registers a :class:`.Subsystem` to this Scheduler, so that the
        Scheduler might know if a default Command needs to be
        run. All :class:`.Subsystem` objects should call this.

        :param system: the system
        """
        if system is not None:
            self.subsystems.add(system)

    def remove(self, command):
        """Removes the :class:`.Command` from the Scheduler.

        :param command: the command to remove
        """
        if command is None or command not in self.commandTable:
            return
        del self.commandTable[command]
        for reqt in command.getRequirements():
            reqt.setCurrentCommand(None)
        command.removed()

    def removeAll(self):
        """Removes all commands
        """
        # TODO: Confirm that this works with "uninteruptible" commands
        for command in self.commandTable:
            for reqt in command.getRequirements():
                reqt.setCurrentCommand(None)
            command.removed()
        self.commandTable.clear()

    def disable(self):
        """Disable the command scheduler.
        """
        self.disabled = True

    def enable(self):
        """Enable the command scheduler.
        """
        self.disabled = False

    def initSendable(self, builder):
        builder.setSmartDashboardType("Scheduler")
        self.namesEntry = builder.getEntry("Names")
        self.idsEntry = builder.getEntry("Ids")
        self.cancelEntry = builder.getEntry("Cancel")
        builder.setUpdateTable(self._updateTable)

    def _updateTable(self):
        if not (self.namesEntry is not None and self.idsEntry is not None and self.cancelEntry is not None):
            return

        toCancel = self.cancelEntry.getDoubleArray([])
        if toCancel:
            for command in self.commandTable:
                if float(id(command)) in toCancel:
                    command.cancel()
            self.cancelEntry.setDoubleArray([])

        if self.runningCommandsChanged:
            commands = []
            ids = []

            for command in self.commandTable:
                commands.append(command.getName())
                ids.append(id(command))

            self.namesEntry.setStringArray(commands)
            self.idsEntry.setDoubleArray(ids)



# validated: 2018-01-28 DV 48ae6c954a75 edu/wpi/first/wpilibj/DriverStation.java
# Copyright (c) 2008-2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.

import threading

import enum
import hal
import sys
import traceback

from networktables import NetworkTables

from .motorsafety import MotorSafety
from .timer import Timer

__all__ = ["DriverStation"]

import logging
logger = logging.getLogger('wpilib.ds')

JOYSTICK_UNPLUGGED_MESSAGE_INTERVAL = 1.0


class MatchDataSender:
    def __init__(self):
        self.table = NetworkTables.getTable('FMSInfo')
        self.typeMetadata = self.table.getEntry('.type')
        self.typeMetadata.forceSetString('FMSInfo')
        self.gameSpecificMessage = self.table.getEntry('GameSpecificMessage')
        self.gameSpecificMessage.forceSetString('')
        self.eventName = self.table.getEntry('EventName')
        self.eventName.forceSetString('')
        self.matchNumber = self.table.getEntry('MatchNumber')
        self.matchNumber.forceSetDouble(0)
        self.replayNumber = self.table.getEntry('ReplayNumber')
        self.replayNumber.forceSetDouble(0)
        self.matchType = self.table.getEntry('MatchType')
        self.matchType.forceSetDouble(0)
        self.alliance = self.table.getEntry('IsRedAlliance')
        self.alliance.forceSetBoolean(True)
        self.station = self.table.getEntry('StationNumber')
        self.station.forceSetDouble(0)
        #self.controlWord = self.table.getEntry('FMSControlData')
        #self.controlWord.forceSetDouble(0)


class DriverStation:
    """Provide access to the network communication data to / from the Driver Station."""

    #: The number of joystick ports
    kJoystickPorts = 6

    _station_numbers = {
        hal.AllianceStationID.kBlue1: 1,
        hal.AllianceStationID.kBlue2: 2,
        hal.AllianceStationID.kBlue3: 3,
        hal.AllianceStationID.kRed1: 1,
        hal.AllianceStationID.kRed2: 2,
        hal.AllianceStationID.kRed3: 3,
    }

    class Alliance(enum.IntEnum):
        """The robot alliance that the robot is a part of."""
        Red = 0
        Blue = 1
        Invalid = 2

    class MatchType(enum.IntEnum):
        None_ = 0
        Practice = 1
        Qualification = 2
        Elimination = 3

    @classmethod
    def _reset(cls):
        ds = getattr(cls, 'instance', None)
        if ds is not None:
            ds.release()
            #hal.giveMultiWait(ds.packetDataAvailableSem)
            ds.thread.join()
            del cls.instance

    @classmethod
    def getInstance(cls):
        """Gets the global instance of the DriverStation.

        :returns: :class:`DriverStation`
        """
        try:
            return cls.instance
        except AttributeError:
            cls.instance = None
            cls.instance = cls()
            return cls.instance

    def __init__(self):
        """DriverStation constructor.

        The single DriverStation instance is created statically with the
        instance static member variable, you should never create a
        DriverStation instance.
        """
        if not hasattr(DriverStation, 'instance') or DriverStation.instance is not None:
            raise ValueError("Do not create DriverStation instances, use DriverStation.getInstance() instead")

        self.waitForDataCount = 0
        self.waitForDataCond = threading.Condition()

        self.cacheDataMutex = threading.RLock()

        self.joystickAxes = [hal.JoystickAxes() for _ in range(self.kJoystickPorts)]
        self.joystickPOVs = [hal.JoystickPOVs() for _ in range(self.kJoystickPorts)]
        self.joystickButtons = [hal.JoystickButtons() for _ in range(self.kJoystickPorts)]
        self.matchInfo = hal.MatchInfo(eventName=b'', gameSpecificMessage=b'')

        self.joystickButtonsPressed = [hal.JoystickButtons() for _ in range(self.kJoystickPorts)]
        self.joystickButtonsReleased = [hal.JoystickButtons() for _ in range(self.kJoystickPorts)]

        self.joystickAxesCache = [hal.JoystickAxes() for _ in range(self.kJoystickPorts)]
        self.joystickPOVsCache = [hal.JoystickPOVs() for _ in range(self.kJoystickPorts)]
        self.joystickButtonsCache = [hal.JoystickButtons() for _ in range(self.kJoystickPorts)]
        self.matchInfoCache = hal.MatchInfo(eventName=b'', gameSpecificMessage=b'')

        self.controlWordMutex = threading.RLock()
        self.controlWordCache = hal.ControlWord()
        self.lastControlWordUpdate = 0

        self.matchDataSender = MatchDataSender()

        # vars not initialized in constructor
        self.nextMessageTime = 0.0

        self.threadKeepAlive = True
        self.waitForDataPredicate = False

        self.userInDisabled = False
        self.userInAutonomous = False
        self.userInTeleop = False
        self.userInTest = False

        # Rest of constructor

        self.thread = threading.Thread(target=self._run, name="FRCDriverStation")
        self.thread.daemon = True
        self.thread.start()

    def release(self):
        """Kill the thread."""
        self.threadKeepAlive = False

    @staticmethod
    def reportError(error, printTrace):
        """Report error to Driver Station, and also prints error to `sys.stderr`.
        Optionally appends stack trace to error message.

        :param printTrace: If True, append stack trace to error string
        """
        DriverStation._reportErrorImpl(True, 1, error, printTrace)

    @staticmethod
    def reportWarning(error, printTrace):
        """Report warning to Driver Station, and also prints error to `sys.stderr`.
        Optionally appends stack trace to error message.

        :param printTrace: If True, append stack trace to warning string
        """
        DriverStation._reportErrorImpl(False, 1, error, printTrace)

    @staticmethod
    def _reportErrorImpl(isError, code, error, printTrace):
        traceString = ""
        locString = ""

        if printTrace:
            exc = sys.exc_info()[0]
            stack = traceback.extract_stack()[:-2]  # last one is this func
            if exc is not None:  # i.e. if an exception is present
                # remove call of full_stack, the printed exception
                # will contain the caught exception caller instead
                del stack[-1]
                del stack[-1]

            locString = "%s.%s:%s" % (stack[-1][0], stack[-1][1], stack[-1][2])

            trc = 'Traceback (most recent call last):\n'
            stackstr = trc + ''.join(traceback.format_list(stack))
            if exc is not None:
                stackstr += '  ' + traceback.format_exc().lstrip(trc)
            traceString += ':\n' + stackstr
            logger.exception(error)
        elif isError:
            logger.error(error)
        else:
            logger.warning(error)

        hal.sendError(isError, code, False,
                      error.encode('utf-8'),
                      locString.encode('utf-8'),
                      traceString.encode('utf-8'), True)

    def getStickAxis(self, stick: int, axis: int) -> float:
        """Get the value of the axis on a joystick.

        This depends on the mapping of the joystick connected to the specified
        port.

        :param stick: The joystick port number
        :param axis: The analog axis value to read from the joystick.

        :returns: The value of the axis on the joystick.
        """
        if axis < 0 or axis >= hal.kMaxJoystickAxes:
            raise IndexError("Joystick axis is out of range")
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            joystickAxes = self.joystickAxes[stick]

            if axis >= joystickAxes.count:
                self._reportJoystickUnpluggedWarning("Joystick axis %d on port %d not available, check if controller is plugged in\n" % (axis, stick))
                return 0.0

            return joystickAxes.axes[axis]

    def getStickPOV(self, stick: int, pov: int) -> int:
        """Get the state of a POV on the joystick.

        :param stick: The joystick port number
        :param pov: which POV

        :returns: The angle of the POV in degrees, or -1 if the POV is not
                  pressed.
        """
        if pov < 0 or pov >= hal.kMaxJoystickPOVs:
            raise IndexError("Joystick POV is out of range")
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            joystickPOVs = self.joystickPOVs[stick]

            if pov >= joystickPOVs.count:
                self._reportJoystickUnpluggedWarning("Joystick POV %d on port %d not available, check if controller is plugged in\n" % (pov, stick))
                return -1
            return joystickPOVs.povs[pov]

    def getStickButtons(self, stick: int) -> int:
        """The state of all the buttons on the joystick.

        :param stick: The joystick port number

        :returns: The state of all buttons, as a bit array.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            return self.joystickButtons[stick].buttons

    def getStickButton(self, stick: int, button: int) -> bool:
        """The state of a button on the joystick. Button indexes begin at 1.

        :param stick: The joystick port number
        :param button: The button index, beginning at 1.

        :returns: The state of the button.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            joystickButtons = self.joystickButtons[stick]
            if button > joystickButtons.count:
                self._reportJoystickUnpluggedWarning("Joystick Button %d on port %d not available, check if controller is plugged in" % (button, stick))
                return False
            if button <= 0:
                self._reportJoystickUnpluggedError("Button indexes begin at 1 for WPILib")
                return False
            return ((0x1 << (button - 1)) & joystickButtons.buttons) != 0

    def getStickButtonPressed(self, stick: int, button: int) -> bool:
        """Whether one joystick button was pressed since the last check.
        Button indices begin at 1.

        :param stick: Joystick to read
        :param button: Button index, beginning at 1
        :returns: Whether the joystick button was pressed since the last check
        """
        if button <= 0:
            self._reportJoystickUnpluggedError("Button indexes begin at 1 for WPILib")
            return False
        if not 0 <= stick < self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        if self.joystickButtonsPressed[stick].buttons & 1 << (button - 1):
            self.joystickButtonsPressed[stick].buttons &= ~(1 << (button - 1))
            return True
        return False

    def getStickButtonReleased(self, stick: int, button: int) -> bool:
        """Whether one joystick button was released since the last check.
        Button indices begin at 1.

        :param stick: Joystick to read
        :param button: Button index, beginning at 1
        :returns: Whether the joystick button was released since the last check
        """
        if button <= 0:
            self._reportJoystickUnpluggedError("Button indexes begin at 1 for WPILib")
            return False
        if not 0 <= stick < self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        if self.joystickButtonsReleased[stick].buttons & 1 << (button - 1):
            self.joystickButtonsReleased[stick].buttons &= ~(1 << (button - 1))
            return True
        return False

    def getStickAxisCount(self, stick: int) -> int:
        """Returns the number of axes on a given joystick port.

        :param stick: The joystick port number

        :returns: The number of axes on the indicated joystick
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            return self.joystickAxes[stick].count

    def getStickPOVCount(self, stick: int) -> int:
        """Returns the number of POVs on a given joystick port.

        :param stick: The joystick port number

        :returns: The number of POVs on the indicated joystick
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            return self.joystickPOVs[stick].count

    def getStickButtonCount(self, stick: int) -> int:
        """Gets the number of buttons on a joystick.

        :param stick: The joystick port number

        :returns: The number of buttons on the indicated joystick.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            return self.joystickButtons[stick].count

    def getJoystickIsXbox(self, stick: int) -> bool:
        """Gets the value of isXbox on a joystick.

        :param stick: The joystick port number

        :returns: A boolean that returns the value of isXbox
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            # TODO: Remove this when calling for descriptor on empty stick no longer crashes.
            if 1 > self.joystickButtons[stick].count and 1 > len(self.joystickAxes[stick]):
                self._reportJoystickUnpluggedWarning("WARNING: Joystick on port {} not avaliable, check if controller is "
                                                     "plugged in.\n".format(stick))
                return False

        return hal.getJoystickIsXbox(stick)

    def getJoystickType(self, stick: int) -> int:
        """
        Gets the value of type on a joystick

        :param stick: The joystick port number

        :returns: An integer that returns the value of type.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            # TODO: Remove this when calling for descriptor on empty stick no longer crashes.
            if 1 > self.joystickButtons[stick].count and 1 > len(self.joystickAxes[stick]):
                self._reportJoystickUnpluggedWarning("Joystick on port {} not avaliable, check if controller is "
                                                     "plugged in.\n".format(stick))
                return -1

        return hal.getJoystickType(stick)

    def getJoystickName(self, stick: int) -> str:
        """
        Gets the name of a joystick

        :param stick: The joystick port number

        :returns: The joystick name.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            # TODO: Remove this when calling for descriptor on empty stick no longer crashes.
            if 1 > self.joystickButtons[stick].count and 1 > len(self.joystickAxes[stick].axes):
                self._reportJoystickUnpluggedError("WARNING: Joystick on port {} not avaliable, check if controller is "
                                                   "plugged in.\n".format(stick))
                return ""

        return hal.getJoystickName(stick)

    def getJoystickAxisType(self, stick: int, axis: int) -> int:
        """Returns the types of Axes on a given joystick port.

        :param stick: The joystick port number
        :param axis: The target axis

        :returns: An integer that reports type of axis the axis is reporting to be
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.cacheDataMutex:
            return hal.getJoystickAxisType(stick, axis)

    def isEnabled(self) -> bool:
        """Gets a value indicating whether the Driver Station requires the
        robot to be enabled.

        :returns: True if the robot is enabled, False otherwise.
        """
        with self.controlWordMutex:
            self._updateControlWord(False)
            return self.controlWordCache.enabled != 0 and self.controlWordCache.dsAttached != 0

    def isDisabled(self) -> bool:
        """Gets a value indicating whether the Driver Station requires the
        robot to be disabled.

        :returns: True if the robot should be disabled, False otherwise.
        """
        return not self.isEnabled()

    def isAutonomous(self) -> bool:
        """Gets a value indicating whether the Driver Station requires the
        robot to be running in autonomous mode.

        :returns: True if autonomous mode should be enabled, False otherwise.
        """
        with self.controlWordMutex:
            self._updateControlWord(False)
            return self.controlWordCache.autonomous != 0

    def isOperatorControl(self) -> bool:
        """Gets a value indicating whether the Driver Station requires the
        robot to be running in operator-controlled mode.

        :returns: True if operator-controlled mode should be enabled,
            False otherwise.
        """
        with self.controlWordMutex:
            self._updateControlWord(False)
            return not (self.controlWordCache.autonomous != 0 or self.controlWordCache.test != 0)

    def isTest(self) -> bool:
        """Gets a value indicating whether the Driver Station requires the
        robot to be running in test mode.

        :returns: True if test mode should be enabled, False otherwise.
        """
        with self.controlWordMutex:
            self._updateControlWord(False)
            return self.controlWordCache.test != 0

    def isDSAttached(self) -> bool:
        """Is the driver station attached to the robot?

        :returns: True if the robot is being controlled by a driver station.
        """
        with self.controlWordMutex:
            self._updateControlWord(False)
            return self.controlWordCache.dsAttached != 0

    def isNewControlData(self) -> bool:
        """Gets if a new control packet from the driver station arrived since
        the last time this function was called.

        :returns: True if the control data has been updated since the last
            call.
        """
        return hal.isNewControlData()

    def isFMSAttached(self) -> bool:
        """Gets if the driver station attached to a Field Management System.

        :returns: True if the robot is competing on a field being controlled
            by a Field Management System
        """
        with self.controlWordMutex:
            self._updateControlWord(False)
            return self.controlWordCache.fmsAttached != 0

    def isSysActive(self):
        """
        Gets a value indicating whether the FPGA outputs are enabled. The
        outputs may be disabled if the robot is disabled or e-stopped, the
        watchdog has expired, or if the roboRIO browns out.

        .. deprecated:: 2018.0.0
            Use :meth:`.RobotController.isSysActive`

        :returns: True if the FPGA outputs are enabled.
        """
        return hal.getSystemActive()

    def isBrownedOut(self):
        """
        Check if the system is browned out.

        .. deprecated:: 2018.0.0
            Use :meth:`.RobotController.isBrownedOut`

        :returns: True if the system is browned out.
        """
        return hal.getBrownedOut()

    def getGameSpecificMessage(self) -> str:
        """Get the game specific message.

        :returns: The game specific message
        """
        with self.cacheDataMutex:
            return self.matchInfo.gameSpecificMessage.decode('utf-8')

    def getEventName(self) -> str:
        """Get the event name

        :returns: The event name
        """
        with self.cacheDataMutex:
            return self.matchInfo.eventName.decode('utf-8')

    def getMatchType(self) -> MatchType:
        """Get the match type.

        :returns: The match type
        """
        with self.cacheDataMutex:
            matchType = self.matchInfo.matchType

        return self.MatchType(matchType)

    def getMatchNumber(self) -> int:
        """Get the match number.

        :returns: The match number
        """
        with self.cacheDataMutex:
            return self.matchInfo.matchNumber

    def getReplayNumber(self) -> int:
        """Get the replay number.

        :returns: The replay number
        """
        with self.cacheDataMutex:
            return self.matchInfo.replayNumber

    def getAlliance(self) -> Alliance:
        """Get the current alliance from the FMS.

        :returns: The current alliance
        """
        allianceStationID = hal.getAllianceStation()
        hAid = hal.AllianceStationID

        if allianceStationID in (hAid.kRed1, hAid.kRed2, hAid.kRed3):
            return self.Alliance.Red
        elif allianceStationID in (hAid.kBlue1, hAid.kBlue2, hAid.kBlue3):
            return self.Alliance.Blue
        else:
            return self.Alliance.Invalid

    def getLocation(self) -> int:
        """Gets the location of the team's driver station controls.

        :returns: The location of the team's driver station controls:
            1, 2, or 3
        """
        allianceStationID = hal.getAllianceStation()
        return self._station_numbers.get(allianceStationID, 0)

    def waitForData(self, timeout: float = None) -> bool:
        """Wait for new data or for timeout, which ever comes first.

        If timeout is None, wait for new data only.

        :param timeout: The maximum time in seconds to wait.

        :returns: True if there is new data, otherwise False
        """
        with self.waitForDataCond:
            currentCount = self.waitForDataCount
            signaled = self.waitForDataCond.wait_for(lambda: self.waitForDataCount != currentCount, timeout)
            if not signaled:
                # Return False if a timeout happened
                return False
        return True

    def getMatchTime(self) -> int:
        """Return the approximate match time.

        The FMS does not currently send the official match time to the robots, but
        does send an approximate match time. The value will count down the time
        remaining in the current period (auto or teleop).

        .. warning::

            This is not an official time (so it cannot be used to argue with
            referees or guarantee that a function will trigger before a match ends).

        The Practice Match function of the DS approximates the behaviour seen on the field.

        :returns: Time remaining in current match period (auto or teleop) in seconds
        """
        return hal.getMatchTime()

    def getBatteryVoltage(self):
        """Read the battery voltage.

        .. deprecated:: 2018.0.0
            Use :meth:`.RobotController.getBatteryVoltage`

        :returns: The battery voltage in Volts.
        """
        return hal.getVinVoltage()

    def InDisabled(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting disabled code; if False, leaving
            disabled code
        """
        self.userInDisabled = entering

    def InAutonomous(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting autonomous code; if False, leaving
            autonomous code
        """
        self.userInAutonomous = entering

    def InOperatorControl(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting teleop code; if False, leaving
            teleop code
        """
        self.userInTeleop = entering

    def InTest(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting test code; if False, leaving test
            code
        """
        self.userInTest = entering

    def _sendMatchData(self):
        alliance = hal.getAllianceStation()
        hAid = hal.AllianceStationID
        isRedAlliance = alliance in {hAid.kRed1, hAid.kRed2, hAid.kRed3}
        stationNumber = self._station_numbers.get(alliance, 0)

        with self.cacheDataMutex:
            eventName = self.matchInfo.eventName.decode('utf-8')
            gameSpecificMessage = self.matchInfo.gameSpecificMessage.decode('utf-8')
            matchNumber = self.matchInfo.matchNumber
            replayNumber = self.matchInfo.replayNumber
            matchType = self.matchInfo.matchType

        self.matchDataSender.alliance.setBoolean(isRedAlliance)
        self.matchDataSender.station.setDouble(stationNumber)
        self.matchDataSender.eventName.setString(eventName)
        self.matchDataSender.gameSpecificMessage.setString(gameSpecificMessage)
        self.matchDataSender.matchNumber.setDouble(matchNumber)
        self.matchDataSender.replayNumber.setDouble(replayNumber)
        self.matchDataSender.matchType.setDouble(matchType)
        #self.matchDataSender.controlWord.setDouble(...)

    def _getData(self):
        """Copy data from the DS task for the user.
        If no new data exists, it will just be returned, otherwise
        the data will be copied from the DS polling loop.
        """

        # Get the status of all of the joysticks
        for stick in range(self.kJoystickPorts):
            hal.getJoystickAxes(stick, self.joystickAxesCache[stick])
            hal.getJoystickPOVs(stick, self.joystickPOVsCache[stick])
            hal.getJoystickButtons(stick, self.joystickButtonsCache[stick])

        hal.getMatchInfo(self.matchInfoCache)

        # Force a control word update, to make sure the data is the newest.
        self._updateControlWord(True)

        # lock joystick mutex to swap cache data
        with self.cacheDataMutex:
            for i in range(self.kJoystickPorts):
                self.joystickButtonsPressed[i].buttons |= \
                    ~self.joystickButtons[i].buttons & self.joystickButtonsCache[i].buttons

                self.joystickButtonsReleased[i].buttons |= \
                    self.joystickButtons[i].buttons & ~self.joystickButtonsCache[i].buttons

            # move cache to actual data
            self.joystickAxes, self.joystickAxesCache = self.joystickAxesCache, self.joystickAxes
            self.joystickButtons, self.joystickButtonsCache = self.joystickButtonsCache, self.joystickButtons
            self.joystickPOVs, self.joystickPOVsCache = self.joystickPOVsCache, self.joystickPOVs

            self.matchInfo, self.matchInfoCache = self.matchInfoCache, self.matchInfo

        with self.waitForDataCond:
            self.waitForDataCount += 1
            self.waitForDataCond.notify_all()

        self._sendMatchData()

    def _reportJoystickUnpluggedError(self, message):
        """
        Reports errors related to unplugged joysticks and throttles them so that they don't overwhelm the DS.
        """
        currentTime = Timer.getFPGATimestamp()
        if currentTime > self.nextMessageTime:
            self.reportError(message, False)
            self.nextMessageTime = currentTime + JOYSTICK_UNPLUGGED_MESSAGE_INTERVAL

    def _reportJoystickUnpluggedWarning(self, message):
        """
        Reports errors related to unplugged joysticks and throttles them so that they don't overwhelm the DS.
        """
        currentTime = Timer.getFPGATimestamp()
        if currentTime > self.nextMessageTime:
            self.reportWarning(message, False)
            self.nextMessageTime = currentTime + JOYSTICK_UNPLUGGED_MESSAGE_INTERVAL

    def _run(self):
        """Provides the service routine for the DS polling thread."""
        safetyCounter = 0

        while self.threadKeepAlive:
            hal.waitForDSData()
            self._getData()

            if self.isDisabled():
                safetyCounter = 0

            safetyCounter += 1
            if safetyCounter >= 4:
                MotorSafety.checkMotors()
                safetyCounter = 0

            if self.userInDisabled:
                hal.observeUserProgramDisabled()

            if self.userInAutonomous:
                hal.observeUserProgramAutonomous()

            if self.userInTeleop:
                hal.observeUserProgramTeleop()

            if self.userInTest:
                hal.observeUserProgramTest()

    def _updateControlWord(self, force):
        """Updates the data in the control word cache.

        Updates if the force parameter is set, or if 50ms have passed since the last update.

        :param force: True to force an update to the cache, otherwise update if 50ms have passed.
        """
        now = hal.getFPGATime()
        with self.controlWordMutex:
            if (now - self.lastControlWordUpdate) > 50 or force:
                hal.getControlWord(self.controlWordCache)
                self.lastControlWordUpdate = now



# validated: 2017-12-22 EN f9bece2ffbf7 edu/wpi/first/wpilibj/NidecBrushless.java
# ----------------------------------------------------------------------------
#  Copyright (c) 2017 FIRST. All Rights Reserved.                             
#  Open Source Software - may be modified and shared by FRC teams. The code   
#  must be accompanied by the FIRST BSD license file in the root directory of 
#  the project.                                                               
# ----------------------------------------------------------------------------
import hal
from .digitaloutput import DigitalOutput
from .sendablebase import SendableBase
from .motorsafety import MotorSafety
from .pwm import PWM
from .interfaces.speedcontroller import SpeedController


__all__ = ['NidecBrushless']


class NidecBrushless(SendableBase, MotorSafety, SpeedController):
    """Nidec Brushless Motor"""

    def __init__(self, pwmChannel, dioChannel):
        """
        :param pwmChannel: The PWM channel that the Nidec Brushless controller is attached to.
                0-9 are on-board, 10-19 are on the MXP port
        :param dioChannel: The DIO channel that the Nidec Brushless controller is attached to.
                0-9 are on-board, 10-25 are on the MXP port
        """
        super().__init__()
        self.dio = DigitalOutput(dioChannel)
        self.addChild(self.dio)
        self.dio.setPWMRate(15625)
        self.dio.enablePWM(0.5)

        self.pwm = PWM(pwmChannel)
        self.addChild(self.pwm)

        MotorSafety.__init__(self)
        self.speed = 0.0
        self.isInverted = False
        self.disabled = False

        hal.report(hal.UsageReporting.kResourceType_NidecBrushless, pwmChannel)
        self.setName("Nidec Brushless", pwmChannel)

    def free(self):
        """Free the resources used by this object."""
        super().free()
        self.dio.free()
        self.pwm.free()

    def set(self, speed):
        """ 
        Set the PWM value.
    
        The PWM value is set using a range of -1.0 to 1.0, appropriately scaling the value for the FPGA.

        :param speed: The speed value between -1.0 and 1.0 to set.
        """
        if not self.disabled:
            self.speed = speed
            self.dio.updateDutyCycle(0.5 + 0.5 * (-speed if self.isInverted else speed))
            self.pwm.setRaw(0xffff)
        self.feed()

    def get(self):
        """
        Get the recently set value of the PWM.  

        :returns: The most recently set value for the PWM between -1.0 and 1.0. 
        :rtype: float
        """
        return self.speed

    def setInverted(self, isInverted):
        """
        :type isInverted: bool
        """
        self.isInverted = isInverted

    def getInverted(self):
        """
        :rtype: bool
        """
        return self.isInverted

    def pidWrite(self, output):
        """ 
        Write out the PID value as seen in the PIDOutput base object. 

        :param output: Write out the PWM value as was found in the PIDController
        :type output: float
        """
        self.set(output)

    def stopMotor(self):
        """
        Stop the motor. This is called by the MotorSafetyHelper object
        when it has a timeout for this PWM and needs to stop it from running.
        Calling :meth:`set` will re-enable the motor.
        """
        self.dio.updateDutyCycle(0.5)
        self.pwm.setDisabled()

    def getDescription(self):
        """
        :rtype: str
        """
        return "Nidec %s" % (self.getChannel(),)

    def disable(self):
        """
        Disable the motor.  The :meth:`enable` function must be called to re-enable
        the motor.
        """
        self.disabled = True
        self.dio.updateDutyCycle(0.5)
        self.pwm.setDisabled()

    def enable(self):
        """
        Re-enable the motor after :meth:`disable` has been called.  
        The :meth:`set` function must be called to set a new motor speed.
        """
        self.disabled = False

    def getChannel(self):
        """
        Gets the channel number associated with the object.

        :returns: The channel number.
        :rtype: int
        """
        return self.pwm.getChannel()

    def initSendable(self, builder):
        builder.setSmartDashboardType("Nidec Brushless")
        builder.setSafeState(self.stopMotor)
        builder.addDoubleProperty("Value", self.get, self.set)



# validated: 2017-10-23 TW 9dc1de1d09da edu/wpi/first/wpilibj/drive/Vector2d.java
import math

__all__ = ['Vector2d']

class Vector2d:
    """This is a 2D vector struct that supports basic operations"""
    def __init__(self, x=0.0, y=0.0):
        """Construct a 2D vector

        :param x: x component of the vector
        :param y: y component of the vector
        """
        self.x = x
        self.y = y

    def rotate(self, angle):
        """Rotate a vector in Cartesian space.

        :param angle: Angle in degrees by which to rotate vector counter-clockwise
        """
        angle = math.radians(angle)
        cosA = math.cos(angle)
        sinA = math.sin(angle)

        x = self.x * cosA - self.y * sinA
        y = self.x * sinA + self.y * cosA
        self.x = x
        self.y = y

    def dot(self, vec):
        """Returns dot product of this vector and argument

        :param vec: Vector with which to perform dot product
        :type vec: Vector2d
        """
        return self.x * vec.x + self.y * vec.y

    def magnitude(self):
        """ Returns magnitude of vector"""
        return math.hypot(self.x, self.y)

    def scalarProject(self, vec):
        """Returns scalar projection of this vector onto argument

        :param vec: Vector onto which to project this vector
        :type vec: Vector2d
        :return: scalar projection of this vector onto argument
        """
        return self.dot(vec) / vec.magnitude()
